from typing import Optional, Tuple

import torch
from torch import Tensor


def get_approximate_basis(
    A: Tensor, q: int, niter: Optional[int] = 2, M: Optional[Tensor] = None
) -> Tensor:
    """Return tensor :math:`Q` with :math:`q` orthonormal columns such
    that :math:`Q Q^H A` approximates :math:`A`. If :math:`M` is
    specified, then :math:`Q` is such that :math:`Q Q^H (A - M)`
    approximates :math:`A - M`.

    .. note:: The implementation is based on the Algorithm 4.4 from
              Halko et al, 2009.

    .. note:: For an adequate approximation of a k-rank matrix
              :math:`A`, where k is not known in advance but could be
              estimated, the number of :math:`Q` columns, q, can be
              choosen according to the following criteria: in general,
              :math:`k <= q <= min(2*k, m, n)`. For large low-rank
              matrices, take :math:`q = k + 5..10`.  If k is
              relatively small compared to :math:`min(m, n)`, choosing
              :math:`q = k + 0..2` may be sufficient.

    .. note:: To obtain repeatable results, reset the seed for the
              pseudorandom number generator

    Args::
        A (Tensor): the input tensor of size :math:`(*, m, n)`

        q (int): the dimension of subspace spanned by :math:`Q`
                 columns.

        niter (int, optional): the number of subspace iterations to
                               conduct; ``niter`` must be a
                               nonnegative integer. In most cases, the
                               default value 2 is more than enough.

        M (Tensor, optional): the input tensor's mean of size
                              :math:`(*, 1, n)`.

    References::
        - Nathan Halko, Per-Gunnar Martinsson, and Joel Tropp, Finding
          structure with randomness: probabilistic algorithms for
          constructing approximate matrix decompositions,
          arXiv:0909.4061 [math.NA; math.PR], 2009 (available at
          `arXiv <http://arxiv.org/abs/0909.4061>`_).
    """

    niter = 2 if niter is None else niter
    m, n = A.shape[-2:]
    dtype = _utils.get_floating_dtype(A)
    matmul = _utils.matmul

    R = torch.randn(n, q, dtype=dtype, device=A.device)

    # The following code could be made faster using torch.geqrf + torch.ormqr
    # but geqrf is not differentiable
    A_H = _utils.transjugate(A)
    if M is None:
        Q = torch.linalg.qr(matmul(A, R)).Q
        for i in range(niter):
            Q = torch.linalg.qr(matmul(A_H, Q)).Q
            Q = torch.linalg.qr(matmul(A, Q)).Q
    else:
        M_H = _utils.transjugate(M)
        Q = torch.linalg.qr(matmul(A, R) - matmul(M, R)).Q
        for i in range(niter):
            Q = torch.linalg.qr(matmul(A_H, Q) - matmul(M_H, Q)).Q
            Q = torch.linalg.qr(matmul(A, Q) - matmul(M, Q)).Q

    return Q



from .TrueCompression import GearLlamaForCausalLMNew
from .TrueCompression import GearLlamaForCausalLM

from .Simulated import CompressionConfig
from .Simulated import SimulatedGearLlamaForCausalLM

# from .modeling_llama_h2o import H2OLlamaForCausalLM, LlamaConfig
from .Simulated import SimulatedGearMistralForCausalLM, MistralConfig
from .Simulated import LlamaForCausalLMH2O



from typing import Any, Dict, List, Optional, Tuple
from .cache_utils import Cache
import torch
from .TrueCompressFunction import (
    true_uniform_quantization_compress,
    true_uniform_quantization_decompress,
    true_outlier_quantization_compress,
    true_outlier_quantization_decompress,
    true_gear_compress,
    true_gear_decompress,
)
from .TrueCompressFunction import (
    true_uniform_quantization_compress_batchwise,
    true_uniform_quantization_decompress_batchwise,
    true_outlier_quantization_compress_batchwise,
    true_outlier_quantization_decompress_batchwise,
    true_gear_compress,
    true_gear_decompress_batchwise,
    true_gear_compress_batchwise,
)

compress_function = {
    "uniform": true_uniform_quantization_compress,
    "outlier": true_outlier_quantization_compress,
    "gear": true_gear_compress,
    "uniform_batch": true_uniform_quantization_compress_batchwise,
    "outlier_batch": true_outlier_quantization_compress_batchwise,
    "gear_batch": true_gear_compress_batchwise,
}
decompress_function = {
    "uniform": true_uniform_quantization_decompress,
    "outlier": true_outlier_quantization_decompress,
    "gear": true_gear_decompress,
    "uniform_batch": true_uniform_quantization_decompress_batchwise,
    "outlier_batch": true_outlier_quantization_decompress_batchwise,
    "gear_batch": true_gear_decompress_batchwise,
}


class StreamCompressedUnion:
    def __init__(self, compress_kwargs: Optional[Dict[str, Any]] = None):
        self.quantize_bit = compress_kwargs["quantize_bit"]
        self.compress_mode = compress_kwargs["compress_mode"]
        self.min = None
        self.step = None
        self.left = compress_kwargs["left"]
        self.rank = compress_kwargs["rank"]
        self.loop = compress_kwargs["loop"]
        self.dtype = None
        self.shape = None
        self.is_compressed = False
        self.cache = None
        self.values = None
        self.indices = None
        self.p_base = None
        self.q_base = None
        self.counter = 0
        self.gap = compress_kwargs["streaming_gap"]
        self.cache_shape = None
        self.buffer = None
        self.cache_shape = None

    def set_cache(self, input: torch.Tensor):

        self.shape = input.shape
        # # has_inf = torch.isinf(input)
        # # has_nan = torch.isnan(input)
        # # print(self.counter,has_inf.any(),has_nan.any())
        # if self.counter != 1 and self.counter % self.gap == 0:
        #     # self.cache = input[:,:,0:-self.counter,:]
        #     self.buffer = input[:,:,-self.counter:,:].clone()
        #     del input
        # else:

        #     self.cache = input
        #     # self.cache = input
        #     pass
        # print("set_cache",self.counter)
        # print(self.cache.dtype)
        if self.counter == 0 or self.counter % self.gap == 0:
            self.cache = input
            self.cache_shape = input.shape
            self.buffer = None
        else:
            buffer_token = self.counter % self.gap
            self.buffer = input[:, :, -buffer_token:, :].clone()
            del input

    def get_cache(self):
        if self.counter == 0 or self.counter % self.gap == 0:
            return self.cache
        else:
            return self.decompress(True)

    def compress(self):

        input = self.cache
        self.dtype = input.dtype
        self.is_compressed = True
        if self.counter == 0 or self.counter % self.gap == 0:
            if self.compress_mode == "uniform":
                output, shape, min, step = compress_function[self.compress_mode](
                    input, self.quantize_bit
                )
                self.cache = output
                self.min = min
                self.step = step

            elif self.compress_mode == "outlier":
                output, shape, min, step, values, indices = compress_function[
                    self.compress_mode
                ](input, self.quantize_bit, self.left)
                self.cache = output
                self.min = min
                self.step = step

                self.values = values
                self.indices = indices
            elif self.compress_mode == "gear":
                output, shape, min, step, values, indices, p_base, q_base = (
                    compress_function[self.compress_mode](
                        input, self.quantize_bit, self.left, self.rank, self.loop
                    )
                )
                self.cache = output
                self.min = min
                self.step = step

                self.values = values
                self.indices = indices
                self.p_base = p_base
                self.q_base = q_base
            elif self.compress_mode == "uniform_batch":
                output, shape, min, step = compress_function[self.compress_mode](
                    input, self.quantize_bit
                )
                self.cache = output
                self.min = min
                self.step = step

            elif self.compress_mode == "outlier_batch":
                output, shape, min, step, values, indices = compress_function[
                    self.compress_mode
                ](input, self.quantize_bit, self.left)
                self.cache = output
                self.min = min
                self.step = step

                self.values = values
                self.indices = indices
            elif self.compress_mode == "gear_batch":
                output, shape, min, step, values, indices, p_base, q_base = (
                    compress_function[self.compress_mode](
                        input, self.quantize_bit, self.left, self.rank, self.loop
                    )
                )
                self.cache = output
                self.min = min
                self.step = step

                self.values = values
                self.indices = indices
                self.p_base = p_base
                self.q_base = q_base
            self.buffer = None
        else:
            pass
        # print("compress",self.counter)
        # print(self.cache.dtype)

    def decompress(self, flag=False):
        self.is_compressed = flag
        # print("decompress",self.counter)
        # print(self.cache.dtype)
        if self.compress_mode == "uniform":
            output = decompress_function[self.compress_mode](
                self.cache,
                self.quantize_bit,
                self.cache_shape,
                self.min,
                self.step,
                self.dtype,
            )
        elif self.compress_mode == "outlier":
            output = decompress_function[self.compress_mode](
                self.cache,
                self.quantize_bit,
                self.cache_shape,
                self.min,
                self.step,
                self.dtype,
                self.values,
                self.indices,
            )
        elif self.compress_mode == "gear":
            output = decompress_function[self.compress_mode](
                self.cache,
                self.quantize_bit,
                self.cache_shape,
                self.min,
                self.step,
                self.dtype,
                self.values,
                self.indices,
                self.p_base,
                self.q_base,
            )
        elif self.compress_mode == "uniform_batch":
            output = decompress_function[self.compress_mode](
                self.cache,
                self.quantize_bit,
                self.cache_shape,
                self.min,
                self.step,
                self.dtype,
            )
        elif self.compress_mode == "outlier_batch":
            output = decompress_function[self.compress_mode](
                self.cache,
                self.quantize_bit,
                self.cache_shape,
                self.min,
                self.step,
                self.dtype,
                self.values,
                self.indices,
            )
        elif self.compress_mode == "gear_batch":
            output = decompress_function[self.compress_mode](
                self.cache,
                self.quantize_bit,
                self.cache_shape,
                self.min,
                self.step,
                self.dtype,
                self.values,
                self.indices,
                self.p_base,
                self.q_base,
            )
        # self.clean_cache()
        if self.buffer is not None:
            output = torch.cat([output, self.buffer], dim=2)
        # print(self.counter,output.shape)
        return output

    def clean_cache(self):
        self.is_compressed = False
        self.cache = None
        self.values = None
        self.indices = None
        self.p_base = None
        self.q_base = None
        self.min = None
        self.step = None


class StreamCompressedCache(Cache):
    def __init__(self) -> None:
        self.key_cache: List[torch.Tensor] = []
        self.value_cache: List[torch.Tensor] = []
        self.seen_tokens = (
            0  # Used in `generate` to keep tally of how many tokens the cache has seen
        )

    def increase_idx(self, layer_idx):
        self.key_cache[layer_idx].counter += 1
        self.value_cache[layer_idx].counter += 1

    def __setitem__(
        self, layer_idx: int, key_value_states: Tuple[torch.Tensor, torch.Tensor]
    ):
        """
        Support for backwards-compatible `past_key_value` assignment, e.g. `past_key_value[0] = (key_states,
        value_states)` to update the cache for the first layer.
        """
        key_states, value_states = key_value_states
        self.key_cache[layer_idx], self.value_cache[layer_idx] = (
            key_states,
            value_states,
        )

    def __getitem__(self, layer_idx: int) -> List[Tuple[torch.Tensor]]:
        """
        Support for backwards-compatible `past_key_value` indexing, e.g. `past_key_value[0][0].shape[2]` to get the
        sequence length.
        """
        if layer_idx < len(self):
            return (self.key_cache[layer_idx], self.value_cache[layer_idx])
        else:
            raise KeyError(
                f"Cache only has {len(self)} layers, attempted to access layer with index {layer_idx}"
            )

    def __iter__(self):
        """
        Support for backwards-compatible `past_key_value` iteration, e.g. `for x in past_key_value:` to iterate over
        keys and values
        """
        for layer_idx in range(len(self)):
            yield (self.key_cache[layer_idx], self.value_cache[layer_idx])

    def __len__(self):
        """
        Support for backwards-compatible `past_key_value` length, e.g. `len(past_key_value)`. This value corresponds
        to the number of layers in the model.
        """
        return len(self.key_cache)

    def update(
        self,
        key_states: torch.Tensor,
        value_states: torch.Tensor,
        layer_idx: int,
        compress_kwargs: Optional[Dict[str, Any]] = None,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Updates the cache with the new `key_states` and `value_states` for the layer `layer_idx`.

        Parameters:
            key_states (`torch.Tensor`):
                The new key states to cache.
            value_states (`torch.Tensor`):
                The new value states to cache.
            layer_idx (`int`):
                The index of the layer to cache the states for.
            cache_kwargs (`Dict[str, Any]`, `optional`):
                Additional arguments for the cache subclass. No additional arguments are used in `DynamicCache`.

        Return:
            A tuple containing the updated key and value states.
        """
        # Update the number of seen tokens
        if layer_idx == 0:
            self.seen_tokens += key_states.shape[-2]
        # print(isinstance(key_states, Cache))
        # Update the cache
        if len(self.key_cache) <= layer_idx:
            # apply compress here if needed
            if compress_kwargs is not None:

                key_union = StreamCompressedUnion(compress_kwargs)
                value_union = StreamCompressedUnion(compress_kwargs)
                key_union.set_cache(key_states)
                value_union.set_cache(value_states)
                self.key_cache.append(key_union)
                self.value_cache.append(value_union)
            else:
                self.key_cache.append(key_states)
                self.value_cache.append(value_states)
        else:
            if compress_kwargs is not None:
                key_union = self.key_cache[layer_idx]
                value_union = self.value_cache[layer_idx]
                if key_union.is_compressed and value_union.is_compressed:
                    previous_key = key_union.decompress()
                    previous_value = value_union.decompress()
                key_union.set_cache(torch.cat([previous_key, key_states], dim=-2))
                value_union.set_cache(torch.cat([previous_value, value_states], dim=-2))
            else:
                self.key_cache[layer_idx] = torch.cat(
                    [self.key_cache[layer_idx], key_states], dim=-2
                )
                self.value_cache[layer_idx] = torch.cat(
                    [self.value_cache[layer_idx], value_states], dim=-2
                )
        if compress_kwargs is not None:
            return key_union.get_cache(), value_union.get_cache()
        else:
            return self.key_cache[layer_idx], self.value_cache[layer_idx]

    def compress(self, layer_idx: int):
        if len(self.key_cache) <= layer_idx:
            return
        key_union = self.key_cache[layer_idx]
        value_union = self.value_cache[layer_idx]
        if not key_union.is_compressed and not value_union.is_compressed:
            key_union.compress()
            value_union.compress()

    def get_seq_length(self, layer_idx: Optional[int] = 0) -> int:
        """Returns the sequence length of the cached states. A layer index can be optionally passed."""
        if len(self.key_cache) <= layer_idx:
            return 0
        return self.key_cache[layer_idx].shape[-2]

    def get_max_length(self) -> Optional[int]:
        """Returns the maximum sequence length of the cached states. DynamicCache does not have a maximum length."""
        return None

    def reorder_cache(self, beam_idx: torch.LongTensor):
        """Reorders the cache for beam search, given the selected beam indices."""
        for layer_idx in range(len(self.key_cache)):
            device = self.key_cache[layer_idx].device
            self.key_cache[layer_idx] = self.key_cache[layer_idx].index_select(
                0, beam_idx.to(device)
            )
            device = self.value_cache[layer_idx].device
            self.value_cache[layer_idx] = self.value_cache[layer_idx].index_select(
                0, beam_idx.to(device)
            )

    def to_legacy_cache(self) -> Tuple[Tuple[torch.Tensor], Tuple[torch.Tensor]]:
        """Converts the `DynamicCache` instance into the its equivalent in the legacy cache format."""
        legacy_cache = ()
        for layer_idx in range(len(self)):
            legacy_cache += ((self.key_cache[layer_idx], self.value_cache[layer_idx]),)
        return legacy_cache

    @classmethod
    def from_legacy_cache(
        cls, past_key_values: Optional[Tuple[Tuple[torch.FloatTensor]]] = None
    ) -> "DynamicCache":
        """Converts a cache in the legacy cache format into an equivalent `DynamicCache`."""
        cache = cls()
        if past_key_values is not None:
            for layer_idx in range(len(past_key_values)):

                key_states, value_states = past_key_values[layer_idx]
                cache.update(key_states, value_states, layer_idx)
        return cache



class CompressionConfig(dict):
    def __init__(
        self,
        compress_method=None,
        attention_number=12,
        quantize_bit=0,
        group_num=0,
        rank=0.0,
        rankv=0.0,
        loop=0,
        top_k=0.0,
        left=0.0,
        stage=1,
        device_num=0,
        batch_num=1,
        start_saving=0,
        locality_saving=0,
        token_preserving=False,
        streaming=False,
        streaming_gap=0,
        iter=0,
        # h2o setings
        heavy_size=0,
        recent_size=0,
    ):
        self.compress_method = compress_method
        self.quantize_bit = quantize_bit
        self.group_num = group_num
        self.rank = rank
        self.rankv = rankv
        self.ranv = rankv
        self.loop = loop
        self.device_num = device_num
        self.attention_number = attention_number
        self.top_k = top_k
        self.left = left
        self.batch_num = batch_num
        self.stage = stage
        self.start_saving = start_saving
        self.locality_saving = locality_saving
        self.token_preserving = token_preserving
        self.iter = iter
        self.heavy_size = heavy_size
        self.recent_size = recent_size
        self.streaming = streaming
        self.streaming_gap = streaming_gap

    def create_attention_config(self, config):
        attention_config = []
        for i in range(self.attention_number):
            attention_config.append(config)
        return attention_config

    def copy_for_all_attention(self):
        self.compress_method = self.create_attention_config(self.compress_method)
        self.quantize_bit = self.create_attention_config(self.quantize_bit)
        self.group_num = self.create_attention_config(self.group_num)
        self.rank = self.create_attention_config(self.rank)
        self.loop = self.create_attention_config(self.loop)
        self.top_k = self.create_attention_config(self.top_k)
        self.device_num = self.create_attention_config(self.device_num)
        self.left = self.create_attention_config(self.left)
        self.stage = self.create_attention_config(self.stage)
        self.rankv = self.create_attention_config(self.rankv)
        self.start_saving = self.create_attention_config(self.start_saving)
        self.locality_saving = self.create_attention_config(self.locality_saving)
        self.token_preserving = self.create_attention_config(self.token_preserving)
        self.iter = self.create_attention_config(self.iter)
        self.heavy_size = self.create_attention_config(self.heavy_size)
        self.recent_size = self.create_attention_config(self.recent_size)
        self.streaming = self.create_attention_config(self.streaming)
        self.streaming_gap = self.create_attention_config(self.streaming_gap)

    def compress_ratio(
        self,
        compress_method,
        seqlen,
        model_dim,
        rank=0,
        rankv=0,
        quantize_bit=0,
        top_k=0,
        left=0.0,
        stage=1,
        batch_num=1,
    ):
        if compress_method == None:
            return 1.0
        elif compress_method == "Picache":
            if seqlen > rank and seqlen > rankv:
                return (
                    2
                    * seqlen
                    * batch_num
                    * model_dim
                    / (
                        ((model_dim + seqlen * batch_num) * (rank + rankv))
                        * quantize_bit
                        / 16
                    )
                )
            elif seqlen <= rank:
                return (
                    (
                        2
                        * seqlen
                        * batch_num
                        * model_dim
                        / (
                            (model_dim + seqlen * batch_num) * rankv
                            + seqlen * batch_num * model_dim
                        )
                    )
                    * 16
                    / quantize_bit
                )

            elif seqlen <= rankv:
                return (
                    (
                        2
                        * seqlen
                        * batch_num
                        * model_dim
                        / (
                            (model_dim + seqlen * batch_num) * rank
                            + seqlen * batch_num * model_dim
                        )
                    )
                    * 16
                    / quantize_bit
                )
        elif compress_method == "poweriteration":
            return (
                seqlen
                * batch_num
                * model_dim
                / ((model_dim + seqlen * batch_num) * rank)
            )
        elif compress_method == "stagept":
            return (
                seqlen
                * batch_num
                * model_dim
                / (model_dim * rank + seqlen * batch_num * (rank / stage))
            )
        elif (
            compress_method == "uniformquantization"
            or compress_method == "groupquantization"
            or compress_method == "sortquantization"
        ):
            return 16 / quantize_bit
        elif compress_method == "pruning":
            return 1 / top_k
        elif (
            compress_method == "densesparseuniformquantization"
            or compress_method == "densesparsesortquantization"
        ):
            return 1 / (quantize_bit / 16 + left)
        elif compress_method == "pt+outlier":
            return (
                seqlen
                * batch_num
                * model_dim
                * 16
                / quantize_bit
                / ((model_dim + seqlen * batch_num) * rank)
            )

    def calculate_compress_ratio_list(self, seqlen, model_dim):
        self.compress_ratio_list = []
        for i, compress_method in enumerate(self.compress_method):
            if compress_method == None:
                self.compress_ratio_list.append(
                    self.compress_ratio(compress_method, seqlen, model_dim)
                )
            elif compress_method == "Picache":
                self.compress_ratio_list.append(
                    self.compress_ratio(
                        compress_method,
                        seqlen,
                        model_dim,
                        rank=self.rank[i],
                        rankv=self.rankv[i],
                        quantize_bit=self.quantize_bit[i],
                        batch_num=self.batch_num,
                        left=self.left[i],
                    )
                )
            elif compress_method == "poweriteration":
                self.compress_ratio_list.append(
                    self.compress_ratio(
                        compress_method,
                        seqlen,
                        model_dim,
                        rank=self.rank[i],
                        batch_num=self.batch_num,
                    )
                )
            elif compress_method == "stagept":
                self.compress_ratio_list.append(
                    self.compress_ratio(
                        compress_method,
                        seqlen,
                        model_dim,
                        rank=self.rank[i],
                        batch_num=self.batch_num,
                        stage=self.stage[i],
                    )
                )
            elif (
                compress_method == "uniformquantization"
                or compress_method == "groupquantization"
                or compress_method == "sortquantization"
            ):
                self.compress_ratio_list.append(
                    self.compress_ratio(
                        compress_method,
                        seqlen,
                        model_dim,
                        rank=0,
                        quantize_bit=self.quantize_bit[i],
                    )
                )
            elif compress_method == "pruning":
                self.compress_ratio_list.append(
                    self.compress_ratio(
                        compress_method,
                        seqlen,
                        model_dim,
                        top_k=self.top_k[i],
                    )
                )
            elif compress_method == "densesparseuniformquantization":
                self.compress_ratio_list.append(
                    self.compress_ratio(
                        compress_method,
                        seqlen,
                        model_dim,
                        quantize_bit=self.quantize_bit[i],
                        left=self.left[i],
                    )
                )
            elif compress_method == "densesparsesortquantization":
                self.compress_ratio_list.append(
                    self.compress_ratio(
                        compress_method,
                        seqlen,
                        model_dim,
                        quantize_bit=self.quantize_bit[i],
                        left=self.left[i],
                    )
                )
            elif compress_method == "pt+outlier":
                self.compress_ratio_list.append(
                    self.compress_ratio(
                        compress_method,
                        seqlen,
                        model_dim,
                        rank=self.rank[i],
                        quantize_bit=self.quantize_bit[i],
                        batch_num=self.batch_num,
                        left=self.left[i],
                    )
                )

    def calculate_compress_ratio_total(self):
        return sum(self.compress_ratio_list) / len(self.compress_ratio_list)

    def __str__(self):
        return f"compress_method:{self.compress_method},\nquantize_bit:{self.quantize_bit},\nrank:{self.rank},\nloop:{self.loop},\ndevice_num:{self.device_num},\ncompressratio:{self.compress_ratio_list},\ncompressratio_total:{self.calculate_compress_ratio_total()}"



import torch
import time


def transfer_8bit_to_4bit(input: torch.Tensor):
    # shape
    assert input.dtype == torch.uint8
    assert input.shape[-1] % 2 == 0
    size = input.shape
    size = size[-1]
    size = int(size / 2)
    input[0:size] = input[0:size] + input[size:] * pow(2, 4)
    cache = input[0:size].clone()
    del input
    return cache


def transfer_4bit_to_8bit(input: torch.Tensor):
    # shape
    assert input.dtype == torch.uint8
    size = input.shape
    low_end = input % pow(2, 4)
    high_end = (input - low_end) / pow(2, 4)
    output = torch.cat((low_end, high_end), dim=0)
    return output


def transfer_8bit_to_4bit_batchwise(input: torch.Tensor):
    # shape
    assert input.dtype == torch.uint8
    assert input.shape[-1] % 2 == 0
    size = input.shape
    size = size[-1]
    size = int(size / 2)
    input[..., 0:size] = input[..., 0:size] + input[..., size:] * pow(2, 4)
    cache = input[..., 0:size].clone()
    # del input
    return cache


def transfer_4bit_to_8bit_batchwise(input: torch.Tensor):
    # shape
    assert input.dtype == torch.uint8
    size = input.shape
    low_end = input % pow(2, 4)
    high_end = (input - low_end) / pow(2, 4)
    output = torch.cat((low_end, high_end), dim=-1)
    return output


def true_uniform_quantization_compress(input: torch.Tensor, quantize_bit):
    if quantize_bit != 8 and quantize_bit != 4:
        raise ValueError("quantize_bit should be 8 or 4")
    shape = input.shape
    bsz = shape[0]
    input = input.reshape(-1)
    if quantize_bit == 8:
        input = input.float()  # convert to 32bits to avoid max - min = inf
    min, max = input.min(), input.max()
    # step = (max - min) / (pow(2, quantize_bit) - 1)
    scale = (max - min) / (2**quantize_bit - 1)
    # print("before min max:",min,max,step)
    quantized_input = (input - min) / scale
    # print("after min max:",quantized_input.min(),quantized_input.max())
    # print("quantized isnan:",torch.any(torch.isnan(quantized_input)))
    quantized_input = quantized_input.round_()
    quantized_input = quantized_input.to(torch.uint8)
    if quantize_bit == 4:
        quantized_input = transfer_8bit_to_4bit(quantized_input)
    # print("isnan:",torch.any(torch.isnan(returning_input)))
    # while(True):
    #     pass
    return quantized_input, shape, min, scale


def true_uniform_quantization_decompress(
    input: torch.Tensor, quantize_bit, shape, min, step, dtype
):
    if quantize_bit != 8 and quantize_bit != 4:
        raise ValueError("quantize_bit should be 8 or 4")
    input = input.reshape(-1)
    if quantize_bit == 8:
        input = input.float()
        input = input * step + min
        output = input.reshape(shape).type(dtype)
    elif quantize_bit == 4:
        input = transfer_4bit_to_8bit(input)

        input = input.type(dtype)
        input = input * step + min
        output = input.reshape(shape)
    return output


def true_outlier_quantization_compress(input: torch.Tensor, quantize_bit, left):
    shape = input.shape
    input = input.reshape(-1)
    left_num = int(len(input) * left / 2)
    value1, indices1 = torch.topk(input, left_num, largest=False)
    value2, indices2 = torch.topk(input, left_num, largest=True)
    values = torch.cat((value1, value2), dim=0)
    indices = torch.cat((indices1, indices2), dim=0)

    input = input.index_fill_(0, indices, 0)
    output, _, min, step = true_uniform_quantization_compress(input, quantize_bit)

    return output, shape, min, step, values, indices


def true_outlier_quantization_decompress(
    input: torch.Tensor, quantize_bit, shape, min, step, dtype, values, indices
):
    input = true_uniform_quantization_decompress(
        input, quantize_bit, shape, min, step, dtype
    )
    input = input.reshape(-1)
    input[indices] = values
    input = input.reshape(shape)
    return input


def fake_quant_error_simulation(input: torch.Tensor, quantize_bit):
    input = input.reshape(-1)

    min, max = input.min(), input.max()
    step = (max - min) / (pow(2, quantize_bit) - 1)
    # print("before min max:",min,max,step)
    error = input - torch.round((input - min) / step)
    return error, min, step


def true_poweriteration(input: torch.Tensor, loop, rank, p_base=None, q_base=None):
    # input size [batch,num_head,seq_len,model_dim/num_head]
    # -> [batch,seq_len,model_dim] -> [batch * seq_len,model_dim]
    # p_base = torch.rand(input.shape[3] * input.shape[1], rank).to(device)
    # q_base = torch.rand(input.shape[0] * input.shape[2], rank).to(device)
    batch, num_head, seq_len, sep_dim = input.shape
    input = (
        input.permute(0, 2, 1, 3).contiguous().view(batch, seq_len, sep_dim * num_head)
    )  # convert to 32bits for qr decomposition
    input = input.view(batch, seq_len, sep_dim * num_head)
    input = input.float()
    if q_base is not None and p_base is not None:
        p_base[0] = p_base[0].float()
        q_base[0] = q_base[0].float()
    else:
        p_base = [torch.rand(batch, sep_dim * num_head, rank).to(input.device).float()]
        q_base = [torch.rand(batch, seq_len, rank).to(input.device).float()]
    # 3 calculation = loop * (matmul) + 2 * qrO(n^2)
    for i in range(loop):
        if i == loop - 1:
            p_base[0] = torch.linalg.qr(p_base[0].float()).Q
        q_base[0] = input @ p_base[0]
        if i == loop - 1:
            q_base[0] = torch.linalg.qr(q_base[0].float()).Q
        p_base[0] = torch.transpose(input, 1, 2) @ q_base[0]
    # input = q_base[0] @ torch.transpose(p_base[0], 0, 1)
    # input = input.view(batch, seq_len, num_head, sep_dim)
    # input = input.permute(0, 2, 1, 3)
    # input = input.type(torch.bfloat16)
    p_base[0] = p_base[0].half()
    q_base[0] = q_base[0].half()
    return p_base, q_base


def true_poweriteration_quantized(
    input: torch.Tensor, loop, rank, p_base=None, q_base=None
):
    # input size [batch,num_head,seq_len,model_dim/num_head]
    # -> [batch,seq_len,model_dim] -> [batch * seq_len,model_dim]
    # p_base = torch.rand(input.shape[3] * input.shape[1], rank).to(device)
    # q_base = torch.rand(input.shape[0] * input.shape[2], rank).to(device)
    batch, num_head, seq_len, sep_dim = input.shape
    input = (
        input.permute(0, 2, 1, 3).contiguous().view(batch, seq_len, sep_dim * num_head)
    )  # convert to 32bits for qr decomposition
    input = input.view(batch, seq_len, sep_dim * num_head)
    input = input.float()
    if q_base is not None and p_base is not None:
        p_base[0] = p_base[0].float()
        q_base[0] = q_base[0].float()
    else:
        p_base = [torch.rand(batch, sep_dim * num_head, rank).to(input.device).float()]
        q_base = [torch.rand(batch, seq_len, rank).to(input.device).float()]
    # 3 calculation = loop * (matmul) + 2 * qrO(n^2)
    for i in range(loop):
        if i == loop - 1:
            p_base[0] = torch.linalg.qr(p_base[0].float()).Q
        q_base[0] = input @ p_base[0]
        if i == loop - 1:
            q_base[0] = torch.linalg.qr(q_base[0].float()).Q
        p_base[0] = torch.transpose(input, 1, 2) @ q_base[0]
    # input = q_base[0] @ torch.transpose(p_base[0], 0, 1)
    # input = input.view(batch, seq_len, num_head, sep_dim)
    # input = input.permute(0, 2, 1, 3)
    # input = input.type(torch.bfloat16)

    # p_base[0] = p_base[0].half()
    # q_base[0] = q_base[0].half()
    #### compress p and q base to 8bits
    p_base[0], shape_p, min_p, scale_p = true_uniform_quantization_compress(
        p_base[0], 8
    )
    q_base[0], shape_q, min_q, scale_q = true_uniform_quantization_compress(
        q_base[0], 8
    )
    return p_base, q_base, shape_p, shape_q, min_p, min_q, scale_p, scale_q


def true_gear_compress(input: torch.Tensor, quantize_bit, left, rank, loop):
    shape = input.shape
    input = input.reshape(-1)
    left_num = int(len(input) * left / 2)
    value1, indices1 = torch.topk(input, left_num, largest=False)
    value2, indices2 = torch.topk(input, left_num, largest=True)
    values = torch.cat((value1, value2), dim=0)
    indices = torch.cat((indices1, indices2), dim=0)
    input = input.index_fill_(0, indices, 0)
    error, min, step = fake_quant_error_simulation(input, quantize_bit)
    error = error.index_fill_(0, indices, 0)
    error = error.reshape(shape)
    p_base, q_base = true_poweriteration(error, loop, rank)
    # has_inf = torch.isinf(p_base[0])
    # has_nan = torch.isnan(p_base[0])
    # if has_inf.any() or has_nan.any():
    #     print("pbase",has_inf.any(),has_nan.any())
    # has_inf = torch.isinf(q_base[0])
    # has_nan = torch.isnan(q_base[0])
    # if has_inf.any() or has_nan.any():
    #     print("qbase",has_inf.any(),has_nan.any())
    output, _, min, step = true_uniform_quantization_compress(input, quantize_bit)
    return output, shape, min, step, values, indices, p_base, q_base


def true_gear_decompress(
    input: torch.Tensor,
    quantize_bit,
    shape,
    min,
    step,
    dtype,
    values,
    indices,
    p_base,
    q_base,
):

    input = true_uniform_quantization_decompress(
        input, quantize_bit, shape, min, step, dtype
    )
    input = input.reshape(-1)
    input[indices] = values
    input = input.reshape(shape)
    error = q_base[0] @ torch.transpose(p_base[0], 1, 2)
    batch, num_head, seq_len, sep_dim = input.shape
    error = error.reshape(batch, seq_len, num_head, sep_dim)
    # error = error.permute(0, 2, 1, 3).type(input.dtype)
    error = error.permute(0, 2, 1, 3)
    input = input + error

    return input


def true_uniform_quantization_compress_batchwise(input: torch.Tensor, quantize_bit):

    if quantize_bit != 8 and quantize_bit != 4:
        raise ValueError("quantize_bit should be 8 or 4")
    shape = input.shape
    bsz = shape[0]
    input = input.reshape(bsz, -1)
    if quantize_bit == 8:
        input = input.float()  # convert to 32bits to avoid max - min = inf
    min, max = input.min(dim=-1).values, input.max(dim=-1).values
    step = (max - min) / (pow(2, quantize_bit) - 1)
    min = min.unsqueeze(1)  # Expand min tensor shape to (bsz, 1)
    step = step.unsqueeze(1)  # Expand step tensor shape to (bsz, 1)
    # print("before min max:",min,max,step)
    input = torch.round((input - min) / step)
    # print("after min max:",quantized_input.min(),quantized_input.max())
    # print("quantized isnan:",torch.any(torch.isnan(quantized_input)))
    input = input.to(torch.uint8)
    if quantize_bit == 4:
        input = transfer_8bit_to_4bit_batchwise(input)
    # print("isnan:",torch.any(torch.isnan(returning_input)))
    # while(True):
    #     pass
    return input, shape, min, step


def true_uniform_quantization_decompress_batchwise(
    input: torch.Tensor, quantize_bit, shape, min, step, dtype
):
    if quantize_bit != 8 and quantize_bit != 4:
        raise ValueError("quantize_bit should be 8 or 4")

    bsz = shape[0]
    input = input.reshape(bsz, -1)
    if quantize_bit == 8:
        input = input.float()
        input = input * step + min

        output = input.reshape(shape).type(dtype)
    elif quantize_bit == 4:
        input = transfer_4bit_to_8bit_batchwise(input)

        input = input.type(dtype)
        input = input * step + min
        output = input.reshape(shape)

    return output


def true_outlier_quantization_compress_batchwise(
    input: torch.Tensor, quantize_bit, left
):
    shape = input.shape
    bsz = shape[0]
    input = input.reshape(bsz, -1)
    left_num = int(input.numel() / bsz * left / 2)

    value1, indices1 = torch.topk(input, left_num, largest=False, dim=-1)
    value2, indices2 = torch.topk(input, left_num, largest=True, dim=-1)

    values = torch.cat((value1, value2), dim=-1)
    indices = torch.cat((indices1, indices2), dim=-1)
    # input = input.index_fill_(0,indices,0)
    # print(indices.shape)
    input.scatter_(1, indices, 0)

    output, _, min, step = true_uniform_quantization_compress(input, quantize_bit)

    return output, shape, min, step, values, indices


def true_outlier_quantization_decompress_batchwise(
    input: torch.Tensor, quantize_bit, shape, min, step, dtype, values, indices
):
    bsz = shape[0]
    input = true_uniform_quantization_decompress(
        input, quantize_bit, shape, min, step, dtype
    )
    input = input.reshape(bsz, -1)
    input.scatter_(1, indices, values)
    input = input.reshape(shape)
    return input


def fake_quant_error_simulation_batchwise(input: torch.Tensor, quantize_bit, bsz):
    input = input.reshape(bsz, -1)

    min, max = input.min(dim=-1).values, input.max(dim=-1).values

    step = (max - min) / (pow(2, quantize_bit) - 1)
    min = min.unsqueeze(1)  # Expand min tensor shape to (bsz, 1)
    step = step.unsqueeze(1)  # Expand step tensor shape to (bsz, 1)
    # print("before min max:",min,max,step)
    error = input - (torch.round((input - min) / step) * step + min)
    return error, min, step


def true_gear_compress_batchwise(input: torch.Tensor, quantize_bit, left, rank, loop):
    shape = input.shape
    bsz = shape[0]
    input = input.reshape(bsz, -1)
    left_num = int(input.numel() / bsz * left / 2)
    value1, indices1 = torch.topk(input, left_num, largest=False, dim=-1)
    value2, indices2 = torch.topk(input, left_num, largest=True, dim=-1)
    values = torch.cat((value1, value2), dim=-1)
    indices = torch.cat((indices1, indices2), dim=-1)
    input = input.scatter_(1, indices, 0.0)
    error, min, step = fake_quant_error_simulation_batchwise(input, quantize_bit, bsz)
    error = error.scatter_(1, indices, 0.0)
    error = error.reshape(shape)
    bsz, num_head, seq_len, sep_dim = shape
    smaller_dim = seq_len if seq_len < sep_dim * num_head else sep_dim * num_head
    rank = int(rank * smaller_dim)
    p_base, q_base = true_poweriteration(error, loop, rank)
    # has_inf = torch.isinf(p_base[0])
    # has_nan = torch.isnan(p_base[0])
    # if has_inf.any() or has_nan.any():
    #     print("pbase",has_inf.any(),has_nan.any())
    # has_inf = torch.isinf(q_base[0])
    # has_nan = torch.isnan(q_base[0])
    # if has_inf.any() or has_nan.any():
    #     print("qbase",has_inf.any(),has_nan.any())
    output, _, min, step = true_uniform_quantization_compress(input, quantize_bit)
    return output, shape, min, step, values, indices, p_base, q_base


def true_gear_decompress_batchwise(
    input: torch.Tensor,
    quantize_bit,
    shape,
    min,
    step,
    dtype,
    values,
    indices,
    p_base,
    q_base,
):
    bsz = shape[0]
    input = true_uniform_quantization_decompress(
        input, quantize_bit, shape, min, step, dtype
    )
    input = input.reshape(bsz, -1)
    input.scatter_(1, indices, values)
    input = input.reshape(shape)
    error = q_base[0] @ torch.transpose(p_base[0], 1, 2)
    batch, num_head, seq_len, sep_dim = input.shape
    error = error.reshape(batch, seq_len, num_head, sep_dim)
    # error = error.permute(0, 2, 1, 3).type(input.dtype)
    error = error.permute(0, 2, 1, 3)
    input = input + error

    return input


def tokenwise_quantization_compress_with_error(input: torch.Tensor, quantize_bit):
    # Currently only support 4 bit quantization
    assert quantize_bit == 4
    shape = input.shape  # bsz, num_head, seq_len, sep_dim
    input = (
        input.permute(0, 2, 1, 3)
        .contiguous()
        .reshape(shape[0], shape[2], shape[1] * shape[3])
    )
    min, max = input.min(dim=-1).values.unsqueeze(-1), input.max(
        dim=-1
    ).values.unsqueeze(-1)
    step = (max - min) / (pow(2, quantize_bit) - 1)

    quantized_input = (input - min) / step
    # quantized_input = F.relu(quantized_input)
    quantized_input = quantized_input.round_()
    error = input - (quantized_input * step + min)

    quantized_input = quantized_input.to(torch.uint8)
    quantized_input = transfer_8bit_to_4bit_batchwise(quantized_input)
    # reshape back to original shape
    # quantized_input = quantized_input.reshape(shape[0],shape[2],shape[1],shape[3])
    error = error.reshape(shape[0], shape[2], shape[1], shape[3])
    # quantized_input = quantized_input.permute(0, 2, 1, 3).contiguous()
    error = error.permute(0, 2, 1, 3).contiguous()
    return quantized_input, error, min, step, shape


def tokenwise_dequantization(
    quantized_input: torch.Tensor, quantize_bit, min, step, shape, dtype
):
    # input size bsz, seq_len, -1
    assert quantize_bit == 4 or quantize_bit == 8
    if quantize_bit == 4:
        quantized_input = transfer_4bit_to_8bit_batchwise(quantized_input)
    quantized_input = quantized_input.to(dtype)
    bsz, num_head, seq_len, sep_dim = shape
    quantized_input = quantized_input * step + min
    quantized_input = (
        quantized_input.reshape(bsz, seq_len, num_head, sep_dim)
        .permute(0, 2, 1, 3)
        .contiguous()
    )

    return quantized_input


def true_gear_tokenwiseQ_compress(input: torch.Tensor, quantize_bit, rank, loop):
    shape = input.shape  # bsz, num_head, seq_len, sep_dim
    bsz = shape[0]
    quantized_input, error, min, step, shape = (
        tokenwise_quantization_compress_with_error(input, quantize_bit)
    )
    # print("min_max_error_compress:",error.min(),error.max())
    bsz, num_head, seq_len, sep_dim = shape
    smaller_dim = seq_len if seq_len < sep_dim * num_head else sep_dim * num_head
    rank = int(rank * smaller_dim)
    p_base, q_base, shape_p, shape_q, min_p, min_q, scale_p, scale_q = (
        true_poweriteration_quantized(error, loop, rank)
    )
    del error
    return (
        quantized_input,
        shape,
        min,
        step,
        p_base,
        q_base,
        shape_p,
        shape_q,
        min_p,
        min_q,
        scale_p,
        scale_q,
    )


def true_gear_tokenwiseQ_decompress(
    quantized_input,
    quantize_bit,
    shape,
    min,
    step,
    p_base,
    q_base,
    shape_p,
    shape_q,
    min_p,
    min_q,
    scale_p,
    scale_q,
    dtype,
):
    # bsz = shape[0]
    #### TODO

    dequantized_input = tokenwise_dequantization(
        quantized_input, quantize_bit, min, step, shape, dtype
    )
    p_base_dequantized = true_uniform_quantization_decompress(
        p_base[0], 8, shape_p, min_p, scale_p, torch.float
    )
    q_base_dequantized = true_uniform_quantization_decompress(
        q_base[0], 8, shape_q, min_q, scale_q, torch.float
    )
    error = q_base_dequantized @ torch.transpose(p_base_dequantized, 1, 2)
    batch, num_head, seq_len, sep_dim = dequantized_input.shape
    error = error.reshape(batch, seq_len, num_head, sep_dim)
    error = error.permute(0, 2, 1, 3)
    # print("min_max_error_decompress:",error.min(),error.max())
    dequantized_input = dequantized_input + error.to(dtype)
    return dequantized_input


#### no pq quantization


def true_gear_tokenwiseQ_compress_nopq(input: torch.Tensor, quantize_bit, rank, loop):
    shape = input.shape  # bsz, num_head, seq_len, sep_dim
    bsz = shape[0]
    quantized_input, error, min, step, shape = (
        tokenwise_quantization_compress_with_error(input, quantize_bit)
    )
    # print("min_max_error_compress:",error.min(),error.max())
    bsz, num_head, seq_len, sep_dim = shape
    smaller_dim = seq_len if seq_len < sep_dim * num_head else sep_dim * num_head
    rank = int(rank * smaller_dim)
    p_base, q_base = true_poweriteration(error, loop, rank)
    del error
    return quantized_input, shape, min, step, p_base, q_base


def true_gear_tokenwiseQ_decompress_nopq(
    quantized_input, quantize_bit, shape, min, step, p_base, q_base, dtype
):
    # bsz = shape[0]
    #### TODO

    dequantized_input = tokenwise_dequantization(
        quantized_input, quantize_bit, min, step, shape, dtype
    )

    error = q_base[0] @ torch.transpose(p_base[0], 1, 2)
    batch, num_head, seq_len, sep_dim = dequantized_input.shape
    error = error.reshape(batch, seq_len, num_head, sep_dim)
    error = error.permute(0, 2, 1, 3)
    # print("min_max_error_decompress:",error.min(),error.max())
    dequantized_input = dequantized_input + error.to(dtype)
    return dequantized_input



from typing import Any, Dict, List, Optional, Tuple

import torch


class Cache:
    """
    Base, abstract class for all caches. The actual data structure is specific to each subclass.
    """

    def update(
        self,
        key_states: torch.Tensor,
        value_states: torch.Tensor,
        layer_idx: int,
        cache_kwargs: Optional[Dict[str, Any]] = None,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Updates the cache with the new `key_states` and `value_states` for the layer `layer_idx`.

        Parameters:
            key_states (`torch.Tensor`):
                The new key states to cache.
            value_states (`torch.Tensor`):
                The new value states to cache.
            layer_idx (`int`):
                The index of the layer to cache the states for.
            cache_kwargs (`Dict[str, Any]`, `optional`):
                Additional arguments for the cache subclass. These are specific to each subclass and allow new types of
                cache to be created.

        Return:
            A tuple containing the updated key and value states.
        """
        raise NotImplementedError("Make sure to implement `update` in a subclass.")

    def get_seq_length(self, layer_idx: Optional[int] = 0) -> int:
        """Returns the sequence length of the cached states. A layer index can be optionally passed."""
        raise NotImplementedError(
            "Make sure to implement `get_seq_length` in a subclass."
        )

    def get_max_length(self) -> Optional[int]:
        """Returns the maximum sequence length of the cached states, if there is any."""
        raise NotImplementedError(
            "Make sure to implement `get_max_length` in a subclass."
        )

    def get_usable_length(
        self, new_seq_length: int, layer_idx: Optional[int] = 0
    ) -> int:
        """Given the sequence length of the new inputs, returns the usable length of the cache."""
        # Cache without size limit -> all cache is usable
        # Cache with size limit -> if the length cache plus the length of the new inputs is larger the maximum cache
        #   length, we will need to evict part of the cache (and thus not all cache is usable)
        max_length = self.get_max_length()
        previous_seq_length = self.get_seq_length(layer_idx)
        if max_length is not None and previous_seq_length + new_seq_length > max_length:
            return max_length - new_seq_length
        return previous_seq_length


class DynamicCache(Cache):
    """
    A cache that grows dynamically as more tokens are generated. This is the default for generative models.

    It stores the Key and Value states as a list of tensors, one for each layer. The expected shape for each tensor is
    `[batch_size, num_heads, seq_len, head_dim]`.
    """

    def __init__(self) -> None:
        self.key_cache: List[torch.Tensor] = []
        self.value_cache: List[torch.Tensor] = []
        self.seen_tokens = (
            0  # Used in `generate` to keep tally of how many tokens the cache has seen
        )

    def __setitem__(
        self, layer_idx: int, key_value_states: Tuple[torch.Tensor, torch.Tensor]
    ):
        """
        Support for backwards-compatible `past_key_value` assignment, e.g. `past_key_value[0] = (key_states,
        value_states)` to update the cache for the first layer.
        """
        key_states, value_states = key_value_states
        self.key_cache[layer_idx], self.value_cache[layer_idx] = (
            key_states,
            value_states,
        )

    def __getitem__(self, layer_idx: int) -> List[Tuple[torch.Tensor]]:
        """
        Support for backwards-compatible `past_key_value` indexing, e.g. `past_key_value[0][0].shape[2]` to get the
        sequence length.
        """
        if layer_idx < len(self):
            return (self.key_cache[layer_idx], self.value_cache[layer_idx])
        else:
            raise KeyError(
                f"Cache only has {len(self)} layers, attempted to access layer with index {layer_idx}"
            )

    def __iter__(self):
        """
        Support for backwards-compatible `past_key_value` iteration, e.g. `for x in past_key_value:` to iterate over
        keys and values
        """
        for layer_idx in range(len(self)):
            yield (self.key_cache[layer_idx], self.value_cache[layer_idx])

    def __len__(self):
        """
        Support for backwards-compatible `past_key_value` length, e.g. `len(past_key_value)`. This value corresponds
        to the number of layers in the model.
        """
        return len(self.key_cache)

    def update(
        self,
        key_states: torch.Tensor,
        value_states: torch.Tensor,
        layer_idx: int,
        cache_kwargs: Optional[Dict[str, Any]] = None,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Updates the cache with the new `key_states` and `value_states` for the layer `layer_idx`.

        Parameters:
            key_states (`torch.Tensor`):
                The new key states to cache.
            value_states (`torch.Tensor`):
                The new value states to cache.
            layer_idx (`int`):
                The index of the layer to cache the states for.
            cache_kwargs (`Dict[str, Any]`, `optional`):
                Additional arguments for the cache subclass. No additional arguments are used in `DynamicCache`.

        Return:
            A tuple containing the updated key and value states.
        """
        # Update the number of seen tokens
        if layer_idx == 0:
            self.seen_tokens += key_states.shape[-2]

        # Update the cache
        if len(self.key_cache) <= layer_idx:
            self.key_cache.append(key_states)
            self.value_cache.append(value_states)
        else:
            self.key_cache[layer_idx] = torch.cat(
                [self.key_cache[layer_idx], key_states], dim=-2
            )
            self.value_cache[layer_idx] = torch.cat(
                [self.value_cache[layer_idx], value_states], dim=-2
            )

        return self.key_cache[layer_idx], self.value_cache[layer_idx]

    def get_seq_length(self, layer_idx: Optional[int] = 0) -> int:
        """Returns the sequence length of the cached states. A layer index can be optionally passed."""
        if len(self.key_cache) <= layer_idx:
            return 0
        return self.key_cache[layer_idx].shape[-2]

    def get_max_length(self) -> Optional[int]:
        """Returns the maximum sequence length of the cached states. DynamicCache does not have a maximum length."""
        return None

    def reorder_cache(self, beam_idx: torch.LongTensor):
        """Reorders the cache for beam search, given the selected beam indices."""
        for layer_idx in range(len(self.key_cache)):
            device = self.key_cache[layer_idx].device
            self.key_cache[layer_idx] = self.key_cache[layer_idx].index_select(
                0, beam_idx.to(device)
            )
            device = self.value_cache[layer_idx].device
            self.value_cache[layer_idx] = self.value_cache[layer_idx].index_select(
                0, beam_idx.to(device)
            )

    def to_legacy_cache(self) -> Tuple[Tuple[torch.Tensor], Tuple[torch.Tensor]]:
        """Converts the `DynamicCache` instance into the its equivalent in the legacy cache format."""
        legacy_cache = ()
        for layer_idx in range(len(self)):
            legacy_cache += ((self.key_cache[layer_idx], self.value_cache[layer_idx]),)
        return legacy_cache

    @classmethod
    def from_legacy_cache(
        cls, past_key_values: Optional[Tuple[Tuple[torch.FloatTensor]]] = None
    ) -> "DynamicCache":
        """Converts a cache in the legacy cache format into an equivalent `DynamicCache`."""
        cache = cls()
        if past_key_values is not None:
            for layer_idx in range(len(past_key_values)):
                key_states, value_states = past_key_values[layer_idx]
                cache.update(key_states, value_states, layer_idx)
        return cache


class SinkCache(Cache):
    """
    A cache that as described in the [Attention Sinks paper](https://arxiv.org/abs/2309.17453). It allows the model to
    generate beyond the length of its context window, without losing fluency in the conversation. As it discards past
    tokens, the model will lose the ability to generate tokens that depend on the context that was discarded.

    It stores the Key and Value states as a list of tensors, one for each layer. The expected shape for each tensor is
    `[batch_size, num_heads, seq_len, head_dim]`.

    Parameters:
        window_length (`int`):
            The length of the context window.
        num_sink_tokens (`int`):
            The number of sink tokens. See the original paper for more information.
    """

    def __init__(self, window_length: int, num_sink_tokens: int) -> None:
        self.key_cache: List[torch.Tensor] = []
        self.value_cache: List[torch.Tensor] = []
        self.window_length = window_length
        self.num_sink_tokens = num_sink_tokens
        self.cos_sin_cache = {}
        self.seen_tokens = (
            0  # Used in `generate` to keep tally of how many tokens the cache has seen
        )

    @staticmethod
    def _rotate_half(x):
        x1 = x[..., : x.shape[-1] // 2]
        x2 = x[..., x.shape[-1] // 2 :]
        return torch.cat((-x2, x1), dim=-1)

    def _apply_key_rotary_pos_emb(
        self, key_states: torch.Tensor, cos: torch.Tensor, sin: torch.Tensor
    ) -> torch.Tensor:
        rotated_key_states = (key_states * cos) + (self._rotate_half(key_states) * sin)
        return rotated_key_states

    def _get_rerotation_cos_sin(
        self, key_states: torch.Tensor, cos: torch.Tensor, sin: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        if key_states.shape[-2] not in self.cos_sin_cache:
            # Upcast to float32 temporarily for better accuracy
            cos = cos.to(torch.float32)
            sin = sin.to(torch.float32)

            # Compute the cos and sin required for back- and forward-rotating to one position earlier in the sequence
            original_cos = cos[self.num_sink_tokens + key_states.shape[-2] :]
            shifted_cos = cos[self.num_sink_tokens : -key_states.shape[-2]]
            original_sin = sin[self.num_sink_tokens + key_states.shape[-2] :]
            shifted_sin = sin[self.num_sink_tokens : -key_states.shape[-2]]
            rerotation_cos = original_cos * shifted_cos + original_sin * shifted_sin
            rerotation_sin = -original_sin * shifted_cos + original_cos * shifted_sin

            self.cos_sin_cache[key_states.shape[-2]] = (
                rerotation_cos.to(key_states.dtype).unsqueeze(0),
                rerotation_sin.to(key_states.dtype).unsqueeze(0),
            )
        return self.cos_sin_cache[key_states.shape[-2]]

    def get_seq_length(self, layer_idx: Optional[int] = 0) -> int:
        """Returns the sequence length of the cached states. A layer index can be optionally passed."""
        # Workaround to make 'key_states.shape[-2] + past_key_value.get_seq_length(self.layer_idx)' <= window_length
        if len(self.key_cache) <= layer_idx:
            return 0
        return self.key_cache[layer_idx].shape[-2]

    def get_max_length(self) -> Optional[int]:
        """Returns the maximum sequence length of the cached states."""
        return self.window_length

    def update(
        self,
        key_states: torch.Tensor,
        value_states: torch.Tensor,
        layer_idx: int,
        cache_kwargs: Optional[Dict[str, Any]] = None,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Updates the cache with the new `key_states` and `value_states` for the layer `layer_idx`.

        Parameters:
            key_states (`torch.Tensor`):
                The new key states to cache.
            value_states (`torch.Tensor`):
                The new value states to cache.
            layer_idx (`int`):
                The index of the layer to cache the states for.
            cache_kwargs (`Dict[str, Any]`, `optional`):
                Additional arguments for the cache subclass. The following arguments can be used in `SinkCache`: `sin`,
                `cos` and `partial_rotation_size`. These arguments are used with models using RoPE, to recompute the
                rotation as the tokens are shifted.

        Return:
            A tuple containing the updated key and value states.
        """
        # Optional kwargs for `SinkCache` -- needed on models using RoPE. `partial_rotation_size` is used on models
        # with partially rotated position embeddings, like Phi or Persimmon.
        sin = cache_kwargs.get("sin")
        cos = cache_kwargs.get("cos")
        partial_rotation_size = cache_kwargs.get("partial_rotation_size")
        using_rope = cos is not None and sin is not None

        # Update the number of seen tokens
        if layer_idx == 0:
            self.seen_tokens += key_states.shape[-2]

        # [bsz, num_heads, seq_len, head_dim]
        if len(self.key_cache) <= layer_idx:
            # Empty cache
            self.key_cache.append(key_states)
            self.value_cache.append(value_states)

        elif key_states.shape[-2] + self.get_seq_length(layer_idx) < self.window_length:
            # Growing cache
            self.key_cache[layer_idx] = torch.cat(
                [self.key_cache[layer_idx], key_states], dim=-2
            )
            self.value_cache[layer_idx] = torch.cat(
                [self.value_cache[layer_idx], value_states], dim=-2
            )

        else:
            # Shifting cache
            keys_to_keep = self.key_cache[layer_idx][
                :,
                :,
                -self.window_length + self.num_sink_tokens + key_states.shape[-2] :,
            ]

            # On RoPE models, we need to recompute the Key rotation as the tokens are shifted
            if using_rope:
                rerotation_cos, rerotation_sin = self._get_rerotation_cos_sin(
                    key_states, cos[: self.window_length], sin[: self.window_length]
                )
                if partial_rotation_size is not None:
                    keys_to_keep, keys_pass = (
                        keys_to_keep[..., :partial_rotation_size],
                        keys_to_keep[..., partial_rotation_size:],
                    )
                keys_to_keep = self._apply_key_rotary_pos_emb(
                    keys_to_keep, rerotation_cos, rerotation_sin
                )
                if partial_rotation_size is not None:
                    keys_to_keep = torch.cat((keys_to_keep, keys_pass), dim=-1)

            # Concatenate sink tokens, shifted & rotated tokens (if needed), and new tokens
            sink_keys = self.key_cache[layer_idx][:, :, : self.num_sink_tokens]
            self.key_cache[layer_idx] = torch.cat(
                [sink_keys, keys_to_keep, key_states], dim=-2
            )

            sink_values = self.value_cache[layer_idx][:, :, : self.num_sink_tokens]
            values_to_keep = self.value_cache[layer_idx][
                :,
                :,
                -self.window_length + self.num_sink_tokens + value_states.shape[-2] :,
            ]
            self.value_cache[layer_idx] = torch.cat(
                [sink_values, values_to_keep, value_states], dim=-2
            )

        return self.key_cache[layer_idx], self.value_cache[layer_idx]

    def reorder_cache(self, beam_idx: torch.LongTensor):
        """Reorders the cache for beam search, given the selected beam indices."""
        for layer_idx in range(len(self.key_cache)):
            device = self.key_cache[layer_idx].device
            self.key_cache[layer_idx] = self.key_cache[layer_idx].index_select(
                0, beam_idx.to(device)
            )
            device = self.value_cache[layer_idx].device
            self.value_cache[layer_idx] = self.value_cache[layer_idx].index_select(
                0, beam_idx.to(device)
            )



from typing import Any, Dict, List, Optional, Tuple
import torch
from .TrueCompressFunction import (
    true_uniform_quantization_compress,
    true_uniform_quantization_decompress,
    true_outlier_quantization_compress,
    true_outlier_quantization_decompress,
    true_gear_compress,
    true_gear_decompress,
    true_gear_tokenwiseQ_compress,
    true_gear_tokenwiseQ_decompress,
    true_gear_tokenwiseQ_compress_nopq,
    true_gear_tokenwiseQ_decompress_nopq,
)
from .TrueCompressFunction import (
    true_uniform_quantization_compress_batchwise,
    true_uniform_quantization_decompress_batchwise,
    true_outlier_quantization_compress_batchwise,
    true_outlier_quantization_decompress_batchwise,
    true_gear_compress,
    true_gear_decompress_batchwise,
    true_gear_compress_batchwise,
)

compress_function = {
    "uniform": true_uniform_quantization_compress,
    "outlier": true_outlier_quantization_compress,
    "gear": true_gear_compress,
    "uniform_batch": true_uniform_quantization_compress_batchwise,
    "outlier_batch": true_outlier_quantization_compress_batchwise,
    "gear_batch": true_gear_compress_batchwise,
    "gear_tokenwiseQ": true_gear_tokenwiseQ_compress,
    "gear_tokenwiseQ_nopq": true_gear_tokenwiseQ_compress_nopq,
}
decompress_function = {
    "uniform": true_uniform_quantization_decompress,
    "outlier": true_outlier_quantization_decompress,
    "gear": true_gear_decompress,
    "uniform_batch": true_uniform_quantization_decompress_batchwise,
    "outlier_batch": true_outlier_quantization_decompress_batchwise,
    "gear_batch": true_gear_decompress_batchwise,
    "gear_tokenwiseQ": true_gear_tokenwiseQ_decompress,
    "gear_tokenwiseQ_nopq": true_gear_tokenwiseQ_decompress_nopq,
}


def detect_infnan(input_tensor, string):
    if torch.isnan(input_tensor).any():
        print(string + "has nan")
        while True:
            pass
    if torch.isinf(input_tensor).any():
        print(string + "has inf")
        while True:
            pass


class CompressUnion:
    def __init__(self, compress_kwargs: Optional[Dict[str, Any]] = None):
        self.quantize_bit = compress_kwargs["quantize_bit"]
        self.compress_mode = compress_kwargs["compress_mode"]
        self.min = None
        self.step = None
        self.min_p = None
        self.min_q = None
        self.step_p = None
        self.step_q = None
        self.left = compress_kwargs["left"]
        self.rank = compress_kwargs["rank"]
        self.loop = compress_kwargs["loop"]
        self.dtype = None
        self.shape = None
        self.shape_p = None
        self.shape_q = None
        self.quantize_part = None
        self.values = None
        self.indices = None
        self.p_base = None
        self.q_base = None
        self.counter = 0
        self.streaming_gap = compress_kwargs["streaming_gap"]
        self.buffer = None
        self.streaming = compress_kwargs["stream"]
        self.seq_length = 0
        self.input_shape = 0

    def compress_function(self, input_tensor: torch.Tensor):
        self.dtype = input_tensor.dtype
        # detect_infnan(input_tensor,"compress input tensor")
        if self.compress_mode == "uniform":
            output, shape, min, step = compress_function[self.compress_mode](
                input_tensor, self.quantize_bit
            )
            self.quantize_part = output
            self.min = min
            self.step = step
            self.shape = shape
        elif self.compress_mode == "outlier":
            output, shape, min, step, values, indices = compress_function[
                self.compress_mode
            ](input_tensor, self.quantize_bit, self.left)
            self.quantize_part = output
            self.min = min
            self.step = step
            self.shape = shape
            self.values = values
            self.indices = indices
        elif self.compress_mode == "gear":
            output, shape, min, step, values, indices, p_base, q_base = (
                compress_function[self.compress_mode](
                    input_tensor, self.quantize_bit, self.left, self.rank, self.loop
                )
            )
            self.quantize_part = output
            self.min = min
            self.step = step
            self.shape = shape
            self.values = values
            self.indices = indices
            self.p_base = p_base
            self.q_base = q_base
        elif self.compress_mode == "uniform_batch":
            output, shape, min, step = compress_function[self.compress_mode](
                input_tensor, self.quantize_bit
            )
            self.quantize_part = output
            self.min = min
            self.step = step
            self.shape = shape
        elif self.compress_mode == "outlier_batch":
            output, shape, min, step, values, indices = compress_function[
                self.compress_mode
            ](input_tensor, self.quantize_bit, self.left)
            self.quantize_part = output
            self.min = min
            self.step = step
            self.shape = shape
            self.values = values
            self.indices = indices
        elif self.compress_mode == "gear_batch":
            output, shape, min, step, values, indices, p_base, q_base = (
                compress_function[self.compress_mode](
                    input_tensor, self.quantize_bit, self.left, self.rank, self.loop
                )
            )
            self.quantize_part = output
            self.min = min
            self.step = step
            self.shape = shape
            self.values = values
            self.indices = indices
            self.p_base = p_base
            self.q_base = q_base
        elif self.compress_mode == "gear_tokenwiseQ":

            (
                quantized_input,
                shape,
                min,
                step,
                p_base,
                q_base,
                shape_p,
                shape_q,
                min_p,
                min_q,
                scale_p,
                scale_q,
            ) = compress_function[self.compress_mode](
                input_tensor, self.quantize_bit, self.rank, self.loop
            )
            self.quantize_part = quantized_input
            self.min = min
            self.step = step
            self.shape = shape
            self.p_base = p_base
            self.q_base = q_base
            self.shape_p = shape_p
            self.shape_q = shape_q
            self.min_p = min_p
            self.min_q = min_q
            self.step_p = scale_p
            self.step_q = scale_q
        elif self.compress_mode == "gear_tokenwiseQ_nopq":
            quantized_input, shape, min, step, p_base, q_base = compress_function[
                self.compress_mode
            ](input_tensor, self.quantize_bit, self.rank, self.loop)
            self.quantize_part = quantized_input
            self.min = min
            self.step = step
            self.shape = shape
            self.p_base = p_base
            self.q_base = q_base
        # print("quantized_part_min_max:",self.quantize_part.min(),self.quantize_part.max(),"p_base_min_max:",self.min_p.min(),self.p_base[0].max(),"q_base_min_max:",self.min_q.min(),self.q_base[0].max())
        # detect_infnan(quantized_input,"compress quantized_input tensor")
        # detect_infnan(self.p_base[0],"compress p_base tensor")
        # detect_infnan(self.q_base[0],"compress q_base tensor")

    def decompress_function(self):
        if self.compress_mode == "uniform":
            output = decompress_function[self.compress_mode](
                self.quantize_part,
                self.quantize_bit,
                self.shape,
                self.min,
                self.step,
                self.dtype,
            )
        elif self.compress_mode == "outlier":
            output = decompress_function[self.compress_mode](
                self.quantize_part,
                self.quantize_bit,
                self.shape,
                self.min,
                self.step,
                self.dtype,
                self.values,
                self.indices,
            )
        elif self.compress_mode == "gear":
            output = decompress_function[self.compress_mode](
                self.quantize_part,
                self.quantize_bit,
                self.shape,
                self.min,
                self.step,
                self.dtype,
                self.values,
                self.indices,
                self.p_base,
                self.q_base,
            )
        elif self.compress_mode == "uniform_batch":
            output = decompress_function[self.compress_mode](
                self.quantize_part,
                self.quantize_bit,
                self.shape,
                self.min,
                self.step,
                self.dtype,
            )
        elif self.compress_mode == "outlier_batch":
            output = decompress_function[self.compress_mode](
                self.quantize_part,
                self.quantize_bit,
                self.shape,
                self.min,
                self.step,
                self.dtype,
                self.values,
                self.indices,
            )
        elif self.compress_mode == "gear_batch":
            output = decompress_function[self.compress_mode](
                self.quantize_part,
                self.quantize_bit,
                self.shape,
                self.min,
                self.step,
                self.dtype,
                self.values,
                self.indices,
                self.p_base,
                self.q_base,
            )
        elif self.compress_mode == "gear_tokenwiseQ":
            output = decompress_function[self.compress_mode](
                self.quantize_part,
                self.quantize_bit,
                self.shape,
                self.min,
                self.step,
                self.p_base,
                self.q_base,
                self.shape_p,
                self.shape_q,
                self.min_p,
                self.min_q,
                self.step_p,
                self.step_q,
                self.dtype,
            )
        elif self.compress_mode == "gear_tokenwiseQ_nopq":
            output = decompress_function[self.compress_mode](
                self.quantize_part,
                self.quantize_bit,
                self.shape,
                self.min,
                self.step,
                self.p_base,
                self.q_base,
                self.dtype,
            )
        # detect_infnan(output,"decompress")
        return output

    def compress(self, input_tensor):
        self.seq_length = input_tensor.shape[-2]
        # print("compress",self.counter)
        self.input_shape = input_tensor.shape
        if self.streaming is True:
            if self.counter % self.streaming_gap == 0:
                self.buffer = None
                self.compress_function(input_tensor)
            else:
                extract_id = self.counter % self.streaming_gap
                self.buffer = input_tensor[:, :, -extract_id:, :].clone()

        else:
            self.compress_function(input_tensor)

    def decompress(self):
        # print("decompress",self.counter)
        if self.streaming is True:
            if self.counter % self.streaming_gap == 0:
                output = self.decompress_function()
                if self.buffer is not None:
                    output = torch.cat([output, self.buffer], dim=-2)

            else:
                output = self.decompress_function()

                output = torch.cat([output, self.buffer], dim=-2)

            self.counter += 1

        else:

            output = self.decompress_function()
        # detect_infnan(output,"decompress output")
        return output



import torch
import time
import torch.nn.functional as F
import torch.nn as nn
import numpy as np


class H2OCache:
    def __init__(self, hh_size):
        self.hh_size = hh_size
        self.h2o_score = torch.zeros(hh_size, hh_size)

    def selection(self, attn_score, key, value, query):
        batch, num_head, seq_len, sep_dim = key.shape
        batch, num_head, inp_len, cache_len = attn_score.shape

        if inp_len == cache_len:
            # prefill get the attn score
            self.h2o_score = attn_score.clone()
        else:
            # update attn_score
            new_score_cache = torch.zeros(batch, num_head, cache_len, cache_len)
            new_score_cache[:, :, :-1, :-1] = self.h2o_score.clone()
            new_score_cache[:, :, :, :-1] = attn_score.clone()
            self.h2o_score = new_score_cache
        if seq_len <= self.hh_size:
            return key, value
        else:
            _, _, cache_len, cache_len = self.h2o_score.shape
            # TODO is this about the last dim or should set dim=-2
            summing_score = torch.sum(self.h2o_score, dim=-1)
            # batch,num_head,cache_len
            summing_score = summing_score.reshape(-1, cache_len)
            choosing_idx = torch.topk(summing_score, self.hh_size, dim=-1).indices
            # pruning the h2o_score
            self.h2o_score = self.h2o_score.view(-1, cache_len, cache_len)
            choosing_idx_flat = choosing_idx.view(-1, self.hh_size).clone()

            choosing_idx_flat1 = choosing_idx_flat.unsqueeze(-1).expand(
                -1, -1, cache_len
            )

            choosing_idx_flat2 = choosing_idx_flat.unsqueeze(-2).expand(
                -1, self.hh_size, -1
            )
            selected_h2o_score1 = torch.gather(self.h2o_score, 1, choosing_idx_flat1)

            selected_h2o_score2 = torch.gather(
                selected_h2o_score1, 2, choosing_idx_flat2
            )
            self.h2o_score = selected_h2o_score2.view(
                batch, num_head, self.hh_size, self.hh_size
            )
            # prun the kv cache
            key = key.reshape(-1, seq_len, sep_dim)
            value = value.reshape(-1, seq_len, sep_dim)
            selected_key = torch.gather(
                key, 1, choosing_idx_flat.unsqueeze(-1).expand(-1, -1, sep_dim)
            )
            selected_value = torch.gather(
                value, 1, choosing_idx_flat.unsqueeze(-1).expand(-1, -1, sep_dim)
            )
            print(
                "selected_key:",
                selected_key.shape,
                "selected_value:",
                selected_value.shape,
                "self.h2o_score:",
                self.h2o_score.shape,
            )
            selected_key = selected_key.reshape(batch, num_head, self.hh_size, sep_dim)
            selected_value = selected_value.reshape(
                batch, num_head, self.hh_size, sep_dim
            )
            batch, num_head, q_seq_len, sep_dim = query.shape

            if q_seq_len > 1:
                # prefill part
                print("prefill")
                query = query.reshape(-1, q_seq_len, sep_dim)
                selected_query = torch.gather(
                    query, 1, choosing_idx_flat.unsqueeze(-1).expand(-1, -1, sep_dim)
                )
                query = query.reshape(batch, num_head, q_seq_len, sep_dim)
            else:
                selected_query = query
            print("query shape:", query.shape)
            return selected_key, selected_value, selected_query

        # else:
        #     # select key value
        #     if inp_len == cache_len:
        #         # prefill get the attn score
        #         self.h2o_score = attn_score.clone()
        #     else:
        # return key, value


def fake_groupwise_asymmetric_quantization(input: torch.Tensor, quantize_bit):
    # what flexgen uses
    batch, num_head, seq_len, sep_dim = input.shape
    dtype = input.dtype
    # 64 groups, group alone sep_dim
    input = input.float()
    input = input.permute(3, 0, 1, 2)
    input = input.reshape(sep_dim, batch * num_head * seq_len)
    min, max = input.min(1).values, input.max(1).values

    quantized_input = input.clone().detach()
    for i in range(len(min)):
        quantized_input[i] = torch.round(
            (input[i] - min[i]) / (max[i] - min[i]) * (pow(2, quantize_bit) - 1)
        )
    for i in range(len(min)):
        input[i] = (
            quantized_input[i] * (max[i] - min[i]) / (pow(2, quantize_bit) - 1) + min[i]
        )
    input = input.reshape(sep_dim, batch, num_head, seq_len)
    input = input.permute(1, 2, 3, 0)
    input = input.type(dtype)
    return input


def fake_groupwise_token_asymmetric_quantization(
    input: torch.Tensor, quantize_bit, group_size=128
):
    batch, num_head, seq_len, sep_dim = input.shape
    dtype = input.dtype
    group_size = 128
    input = (
        input.permute(0, 2, 1, 3).contiguous().view(batch, seq_len, sep_dim * num_head)
    )
    num_groups = (sep_dim * num_head) // group_size
    if num_groups * group_size != input.shape[-1]:
        raise ValueError("group_size should be a factor of the last dimension size")

    input_in_groups = input.view(batch, seq_len, num_groups, group_size)

    mx, mn = input_in_groups.max(dim=-1)[0], input_in_groups.min(dim=-1)[0]
    mx, mn = mx.unsqueeze(-1), mn.unsqueeze(-1)

    scale = (mx - mn) / (2**quantize_bit - 1)
    input_in_groups = (input_in_groups - mn) / scale
    input_in_groups = F.relu(input_in_groups)
    rounded_input_in_groups = input_in_groups.round_()
    dequantized_input_in_groups = rounded_input_in_groups * scale + mn
    dequantized_input = dequantized_input_in_groups.view(
        batch, seq_len, num_head, sep_dim
    )
    dequantized_input = dequantized_input.permute(0, 2, 1, 3)
    dequantized_input = dequantized_input.type(dtype)
    # reshape the input back to its original shape
    input = input.view(batch, seq_len, num_head, sep_dim)
    input = input.permute(0, 2, 1, 3).contiguous().type(dtype)
    return dequantized_input


def fake_groupwise_channel_asymmetric_quantization(
    input: torch.Tensor, quantize_bit, group_size=128
):
    batch, num_head, seq_len, sep_dim = input.shape
    dtype = input.dtype
    group_size = 128
    input = (
        input.permute(0, 2, 1, 3).contiguous().view(batch, seq_len, sep_dim * num_head)
    )

    mx, mn = input.max(dim=-2)[0], input.min(dim=-2)[0]
    mx, mn = mx.unsqueeze(-2), mn.unsqueeze(-2)

    scale = (mx - mn) / (2**quantize_bit - 1)
    quantized_input = (input - mn) / scale
    quantized_input = F.relu(quantized_input)
    rounded_input = quantized_input.round_()
    dequantized_input = rounded_input * scale + mn
    dequantized_input = dequantized_input.view(batch, seq_len, num_head, sep_dim)
    dequantized_input = dequantized_input.permute(0, 2, 1, 3)
    dequantized_input = dequantized_input.type(dtype)
    # reshape the input back to its original shape
    input = input.view(batch, seq_len, num_head, sep_dim)
    input = input.permute(0, 2, 1, 3).contiguous().type(dtype)
    return dequantized_input


def fake_smoothquatization(input, quantize_bit):
    pass


def fake_svd_lowrank2d(input: torch.Tensor, q):
    batch, seq_len, model_dim = input.shape

    input = input.reshape(seq_len, model_dim)
    U, S, V = torch.svd_lowrank(input, q=q)
    V = V.transpose(-1, -2)
    S = torch.diag_embed(S)
    output = torch.matmul(U[..., :, :], S[..., :, :])
    output = torch.matmul(output[..., :, :], V[..., :, :])
    output = output.reshape(batch, seq_len, model_dim)
    return output


def fake_svd_lowrank(input: torch.Tensor, q):
    batch, num_head, seq_len, sep_dim = input.shape

    input = input.permute(0, 2, 1, 3).reshape(batch * seq_len, num_head * sep_dim)
    U, S, V = torch.svd_lowrank(input, q=q)
    V = V.transpose(-1, -2)
    S = torch.diag_embed(S)
    output = torch.matmul(U[..., :, :], S[..., :, :])
    output = torch.matmul(output[..., :, :], V[..., :, :])
    output = output.reshape(batch, seq_len, num_head, sep_dim).permute(0, 2, 1, 3)
    return output


def fake_uniformquantization(input: torch.Tensor, quantize_bit):
    shape = input.shape
    dtype = input.dtype
    input = input.reshape(-1)
    input = input.float()  # convert to 32bits to avoid max - min = inf
    min, max = input.min(), input.max()
    step = (max - min) / (pow(2, quantize_bit) - 1)
    # print("before min max:",min,max,step)
    quantized_input = torch.round((input - min) / step)
    # print("after min max:",quantized_input.min(),quantized_input.max())
    # print("quantized isnan:",torch.any(torch.isnan(quantized_input)))
    dequantized_input = (quantized_input * step) + min
    returning_input = dequantized_input.reshape(shape)
    returning_input = returning_input.type(dtype)
    # print("isnan:",torch.any(torch.isnan(returning_input)))
    # while(True):
    #     pass
    return returning_input


# def fake_uniformquantization_for_clusterquant(input: torch.Tensor, cluster_num):
#     shape = input.shape
#     input = input.reshape(-1)
#     input = input.float()  # convert to 32bits to avoid max - min = inf
#     min, max = input.min(), input.max()
#     # print("before min max:",min,max)
#     step = (max - min) / (cluster_num - 1)
#     quantized_input = torch.floor((input - min) / step)
#     # print("after min max:",quantized_input.min(),quantized_input.max())
#     # print("quantized isnan:",torch.any(torch.isnan(quantized_input)))
#     dequantized_input = (quantized_input * step) + min
#     returning_input = dequantized_input.reshape(shape)
#     returning_input = returning_input.half()
#     # print("isnan:",torch.any(torch.isnan(returning_input)))
#     # while(True):
#     #     pass
#     return returning_input


def fake_top_k_pruning(input_tensor, topk):
    # Flatten the input tensor
    flat_input = input_tensor.reshape(-1)

    # Calculate the absolute values of the flattened input tensor
    abs_values = torch.abs(flat_input)
    k = int(len(abs_values) * topk)
    # Find the top-k largest absolute values and their indices
    _, topk_indices = torch.topk(abs_values, k)

    # Create a mask to keep the top-k values
    mask = torch.zeros_like(flat_input)
    mask[topk_indices] = 1

    # Apply the mask to prune the tensor
    pruned_tensor = flat_input * mask

    # Reshape the pruned tensor to match the original shape
    pruned_tensor = pruned_tensor.view(input_tensor.size())

    return pruned_tensor


def fake_sort_quantization(input_tensor, group_num, quantize_bit):
    shape = input_tensor.shape
    input_tensor = input_tensor.reshape(-1)
    input_tensor = input_tensor.float()  # convert to 32bits to avoid max - min = inf
    sortedtensor, indices = torch.sort(input_tensor)
    group_size = len(sortedtensor) // group_num
    left_values = len(sortedtensor) % group_num

    group_bits = quantize_bit // group_num
    # print(indices)
    for i in range(group_num):
        if i == 0:
            sortedtensor[0 : group_size + left_values] = fake_uniformquantization(
                sortedtensor[0 : group_size + left_values], group_bits
            )
            # print(sortedtensor[0 : group_size + left_values].min(),sortedtensor[0 : group_size + left_values].max())
        else:
            sortedtensor[
                i * group_size + left_values : (i + 1) * group_size + left_values
            ] = fake_uniformquantization(
                sortedtensor[
                    i * group_size + left_values : (i + 1) * group_size + left_values
                ],
                group_bits,
            )
            # print(sortedtensor[i * group_size + left_values : (i + 1) * group_size + left_values].min(),sortedtensor[i * group_size + left_values : (i + 1) * group_size + left_values].max())

    input_tensor[indices] = sortedtensor
    input_tensor = input_tensor.reshape(shape)
    return input_tensor


def fake_dense_sparse_uniformquantization(input: torch.Tensor, quantize_bit, left):
    shape = input.shape
    dtype = input.dtype
    input = input.reshape(-1)
    input = input.float()  # convert to 32bits to avoid max - min = inf
    sortedtensor, indices = torch.sort(input)
    left_num = int(len(sortedtensor) * left / 2)
    sortedtensor[left_num:-left_num] = fake_uniformquantization(
        sortedtensor[left_num:-left_num], quantize_bit
    )
    input[indices] = sortedtensor
    input = input.reshape(shape)
    input = input.type(dtype)
    return input


def fake_dense_sparse_sortquantization(
    input: torch.Tensor, quantize_bit, group_num, left
):
    shape = input.shape
    input = input.reshape(-1)
    input = input.float()  # convert to 32bits to avoid max - min = inf
    sortedtensor, indices = torch.sort(input)
    # this means we will left 10% of the data and not compress them
    left_num = int(len(sortedtensor) * left / 2)

    quantization_list = sortedtensor[left_num:-left_num]
    group_size = len(quantization_list) // group_num
    # these are values that going to be compressed.
    left_values = len(quantization_list) % group_num
    group_bits = quantize_bit // group_num
    print("left values", left_values)
    for i in range(group_num):
        if i == 0:
            quantization_list[0 : group_size + left_values] = fake_uniformquantization(
                quantization_list[0 : group_size + left_values], group_bits
            )
            print(
                quantization_list[0 : group_size + left_values].min(),
                sortedtensor[0 : group_size + left_values].max(),
            )
        else:
            quantization_list[
                i * group_size + left_values : (i + 1) * group_size + left_values
            ] = fake_uniformquantization(
                quantization_list[
                    i * group_size + left_values : (i + 1) * group_size + left_values
                ],
                group_bits,
            )
            print(
                quantization_list[
                    i * group_size + left_values : (i + 1) * group_size + left_values
                ].min(),
                sortedtensor[
                    i * group_size + left_values : (i + 1) * group_size + left_values
                ].max(),
            )
    sortedtensor[left_num:-left_num] = quantization_list

    input[indices] = sortedtensor

    input = input.reshape(shape)
    return input


def fake_poweriteration(input: torch.Tensor, loop, rank, device, p_base, q_base):
    # input size [batch,num_head,seq_len,model_dim/num_head]
    # -> [batch,seq_len,model_dim] -> [batch * seq_len,model_dim]
    # p_base = torch.rand(input.shape[3] * input.shape[1], rank).to(device)
    # q_base = torch.rand(input.shape[0] * input.shape[2], rank).to(device)
    dtype = input.dtype
    batch, num_head, seq_len, sep_dim = input.shape
    input = (
        input.permute(0, 2, 1, 3).contiguous().view(batch, seq_len, sep_dim * num_head)
    )  # convert to 32bits for qr decomposition
    input = input.view(seq_len * batch, sep_dim * num_head)
    input = input.float()
    if q_base is not None and p_base is not None:
        p_base[0] = p_base[0].float()
        q_base[0] = q_base[0].float()
    else:
        p_base = [torch.rand(sep_dim * num_head, rank).to(input.device)]
        q_base = [torch.rand(batch * seq_len, rank).to(input.device)]
    # 3 calculation = loop * (matmul) + 2 * qrO(n^2)
    for i in range(loop):
        if i == loop - 1:
            p_base[0] = torch.linalg.qr(p_base[0]).Q
        q_base[0] = input @ p_base[0]
        if i == loop - 1:
            q_base[0] = torch.linalg.qr(q_base[0]).Q
        p_base[0] = torch.transpose(input, 0, 1) @ q_base[0]
    input = q_base[0] @ torch.transpose(p_base[0], 0, 1)
    input = input.view(batch, seq_len, num_head, sep_dim)
    input = input.permute(0, 2, 1, 3)
    input = input.type(dtype)
    p_base[0] = p_base[0].half()
    q_base[0] = q_base[0].half()
    return input


def fake_poweriteration_with_outlierquant(
    input: torch.Tensor, loop, rank, device, p_base, q_base, quantize_bit, left
):
    batch, num_head, seq_len, sep_dim = input.shape
    input = (
        input.permute(0, 2, 1, 3).contiguous().view(batch, seq_len, sep_dim * num_head)
    )
    input = input.float()
    # convert to 32bits for qr decomposition
    input = input.view(seq_len * batch, sep_dim * num_head)
    # 3 calculation = loop * (matmul) + 2 * qrO(n^2)
    for i in range(loop):
        if i == loop - 1:
            p_base[0] = torch.linalg.qr(p_base[0]).Q
        q_base[0] = input @ p_base[0]
        if i == loop - 1:
            q_base[0] = torch.linalg.qr(q_base[0]).Q
        p_base[0] = torch.transpose(input, 0, 1) @ q_base[0]
    # outlier quantization
    q_base[0] = fake_dense_sparse_uniformquantization(q_base[0], quantize_bit, left)
    p_base[0] = fake_dense_sparse_uniformquantization(p_base[0], quantize_bit, left)
    input = q_base[0] @ torch.transpose(p_base[0], 0, 1)
    input = input.view(batch, seq_len, num_head, sep_dim)
    input = input.permute(0, 2, 1, 3)
    input = input.half()
    return input


def fake_stage_poweriteration(
    input: torch.Tensor, loop, rank, stage, device, p_base, q_base
):
    pass
    # Lets say input is of size x,n,m x= stage, p would be [m,rank] q would be [n,rank], but p is separate to [m,rank/x] ad q is [n1,rank/x], [n2,rank/x], [n3,rank/x]
    batch, num_head, seq_len, sep_dim = input.shape
    input = input.float()
    input = (
        input.permute(0, 2, 1, 3).contiguous().view(batch, seq_len, sep_dim * num_head)
    )  # convert to 32bits for qr decomposition
    # input = input.view(stage,int(batch/stage), seq_len, sep_dim * num_head)
    input = input.view(seq_len * batch, sep_dim * num_head)
    stage_dim = seq_len * batch // stage
    stage_rank = rank // stage
    for i in range(stage):
        if i != stage - 1:
            p_slice = p_base[0][:, i * stage_rank : (i + 1) * stage_rank]
            q_slice = q_base[0][
                i * stage_dim : (i + 1) * stage_dim,
                i * stage_rank : (i + 1) * stage_rank,
            ]
            input_slice = input[i * stage_dim : (i + 1) * stage_dim, :]
        else:
            p_slice = p_base[0][:, i * stage_rank :]
            q_slice = q_base[0][i * stage_dim :, i * stage_rank :]
            input_slice = input[i * stage_dim :, :]
        # print(p_slice.shape,q_slice.shape,input_slice.shape)
        for j in range(loop):
            if j == loop - 1:
                p_slice = torch.linalg.qr(p_slice).Q
            q_slice = input_slice @ p_slice
            if j == loop - 1:
                q_slice = torch.linalg.qr(q_slice).Q
            p_slice = torch.transpose(input_slice, 0, 1) @ q_slice
        input_slice = q_slice @ torch.transpose(p_slice, 0, 1)
        if i != stage - 1:
            input[i * stage_dim : (i + 1) * stage_dim, :] = input_slice
        else:
            input[i * stage_dim :, :] = input_slice
    input = input.view(batch, seq_len, num_head, sep_dim)
    input = input.permute(0, 2, 1, 3)
    input = input.half()
    return input


def ptcompress(cache: torch.Tensor, p_buffer, q_buffer, loop):
    # cache size [batch,num_head,seq_len,model_dim/num_head]
    # -> [batch,seq_len,model_dim] -> [batch * seq_len,model_dim]
    # p_buffer of size [model_dim, rank], q_buffer of size [batch * seq_len,rank]
    batch, num_head, seq_len, sep_dim = cache.shape
    cache = (
        cache.permute(0, 2, 1, 3).contiguous().view(batch, seq_len, sep_dim * num_head)
    )  # convert to 32bits for qr decomposition
    p_buffer[0] = p_buffer[0]
    q_buffer[0] = q_buffer[0]
    cache = cache.view(seq_len * batch, sep_dim * num_head)
    for i in range(loop):
        if i == loop - 1:
            p_buffer[0] = torch.linalg.qr(p_buffer[0]).Q
        q_buffer[0] = cache @ p_buffer[0]
        if i == loop - 1:
            q_buffer[0] = torch.linalg.qr(q_buffer[0]).Q
        p_buffer[0] = torch.transpose(cache, 0, 1) @ q_buffer[0]
    p_buffer[0] = p_buffer[0]
    q_buffer[0] = q_buffer[0]
    cache = cache.view(batch, num_head, seq_len, sep_dim)
    return p_buffer, q_buffer, batch, num_head, seq_len, sep_dim


def ptdecompress(p_buffer, q_buffer, batch, num_head, seq_len, sep_dim):
    # p_buffer of size [model_dim, rank], q_buffer of size [batch * seq_len,rank]
    matrix = q_buffer[0] @ torch.transpose(p_buffer[0], 0, 1)
    matrix = matrix.view(batch, num_head, seq_len, sep_dim)
    return matrix


def fake_picache_compress(
    input: torch.Tensor, loop, rank, device, p_base, q_base, quantize_bit, left
):
    input = input.float()
    p_base[0] = p_base[0].float()
    q_base[0] = q_base[0].float()
    batch, num_head, seq_len, sep_dim = input.shape
    input = (
        input.permute(0, 2, 1, 3).contiguous().view(batch, seq_len, sep_dim * num_head)
    )  # convert to 32bits for qr decomposition
    input = input.view(seq_len * batch, sep_dim * num_head)
    # 3 calculation = loop * (matmul) + 2 * qrO(n^2)
    for i in range(loop):
        if i == loop - 1:
            p_base[0] = torch.linalg.qr(p_base[0]).Q
        q_base[0] = input @ p_base[0]
        if i == loop - 1:
            q_base[0] = torch.linalg.qr(q_base[0]).Q
        p_base[0] = torch.transpose(input, 0, 1) @ q_base[0]
    # outlier quantization
    q_base[0] = fake_dense_sparse_uniformquantization(q_base[0], quantize_bit, left)
    p_base[0] = fake_dense_sparse_uniformquantization(p_base[0], quantize_bit, left)
    input = q_base[0] @ torch.transpose(p_base[0], 0, 1)
    input = input.view(batch, seq_len, num_head, sep_dim)
    input = input.permute(0, 2, 1, 3)
    input = input.half()
    p_base[0] = p_base[0].half()
    q_base[0] = q_base[0].half()
    return input


def compress_insert_function(
    previous_key,
    previous_value,
    compress_config,
    layer_idx,
    pbase1=None,
    qbase1=None,
    pbase2=None,
    qbase2=None,
    attn_weights=None,
):
    batch, num_head, seq_len, sep_dim = previous_key.shape
    if compress_config.token_preserving[layer_idx] == True:
        starting_idx = int(compress_config.start_saving[layer_idx] * seq_len)
        locality_idx = int(compress_config.locality_saving[layer_idx] * seq_len)
    else:
        starting_idx = int(0)
        locality_idx = -seq_len
    # print("starting_idx:", starting_idx, "locality_idx:", locality_idx,compress_config.token_preserving[layer_idx],batch, num_head, seq_len, sep_dim)
 

    if compress_config.compress_method[layer_idx] == "groupquantization":
        previous_key[:, :, starting_idx:-locality_idx, :] = (
            fake_groupwise_asymmetric_quantization(
                previous_key[:, :, starting_idx:-locality_idx, :],
                compress_config.quantize_bit[layer_idx],
            )
        )
        if previous_value is not None:
            previous_value[:, :, starting_idx:-locality_idx, :] = (
                fake_groupwise_asymmetric_quantization(
                    previous_value[:, :, starting_idx:-locality_idx, :],
                    compress_config.quantize_bit[layer_idx],
                )
            )

    # SK: TODO take a look at this: idea merged from KIVI, token grouping does reasonably well on GSM8k-CoT at 4-bit, around ~13%
    if compress_config.compress_method[layer_idx] == "groupquantization_token":
        previous_key[:, :, starting_idx:-locality_idx, :] = (
            fake_groupwise_token_asymmetric_quantization(
                previous_key[:, :, starting_idx:-locality_idx, :],
                compress_config.quantize_bit[layer_idx],
                128,
            )
        )
        if previous_value is not None:
            previous_value[:, :, starting_idx:-locality_idx, :] = (
                fake_groupwise_token_asymmetric_quantization(
                    previous_value[:, :, starting_idx:-locality_idx, :],
                    compress_config.quantize_bit[layer_idx],
                    128,
                )
            )

    # SK: TODO take a look at this: idea merged from KIVI, this channel grouping does better tha what was implemented earlier, Please check the code!
    if compress_config.compress_method[layer_idx] == "groupquantization_channel":
        previous_key[:, :, starting_idx:-locality_idx, :] = (
            fake_groupwise_channel_asymmetric_quantization(
                previous_key[:, :, starting_idx:-locality_idx, :],
                compress_config.quantize_bit[layer_idx],
                128,
            )
        )
        if previous_value is not None:
            previous_value[:, :, starting_idx:-locality_idx, :] = (
                fake_groupwise_channel_asymmetric_quantization(
                    previous_value[:, :, starting_idx:-locality_idx, :],
                    compress_config.quantize_bit[layer_idx],
                    128,
                )
            )

    # SK: TODO take a look at this: idea merged from KIVI
    if compress_config.compress_method[layer_idx] == "groupquantization_kc_vt":
        previous_key[:, :, starting_idx:-locality_idx, :] = (
            fake_groupwise_channel_asymmetric_quantization(
                previous_key[:, :, starting_idx:-locality_idx, :],
                compress_config.quantize_bit[layer_idx],
                128,
            )
        )
        if previous_value is not None:
            previous_value[:, :, starting_idx:-locality_idx, :] = (
                fake_groupwise_token_asymmetric_quantization(
                    previous_value[:, :, starting_idx:-locality_idx, :],
                    compress_config.quantize_bit[layer_idx],
                    128,
                )
            )
    if compress_config.compress_method[layer_idx] == "groupquantization_kc_vt":
        previous_key[:, :, starting_idx:-locality_idx, :] = (
            fake_groupwise_channel_asymmetric_quantization(
                previous_key[:, :, starting_idx:-locality_idx, :],
                compress_config.quantize_bit[layer_idx],
                128,
            )
        )
        if previous_value is not None:
            previous_value[:, :, starting_idx:-locality_idx, :] = (
                fake_groupwise_token_asymmetric_quantization(
                    previous_value[:, :, starting_idx:-locality_idx, :],
                    compress_config.quantize_bit[layer_idx],
                    128,
                )
            )
    if compress_config.compress_method[layer_idx] == "groupquantization_kc_vt_preserving":
        previous_key[:,:,compress_config.start_saving[layer_idx]:-compress_config.locality_saving[layer_idx],:] = (
            fake_groupwise_channel_asymmetric_quantization(
                previous_key[:,:,compress_config.start_saving[layer_idx]:-compress_config.locality_saving[layer_idx],:],
                compress_config.quantize_bit[layer_idx],
                128,
            )
        )
        if previous_value is not None:
            previous_value[:,:,compress_config.start_saving[layer_idx]:-compress_config.locality_saving[layer_idx],:] = (
                fake_groupwise_token_asymmetric_quantization(
                    previous_value[:,:,compress_config.start_saving[layer_idx]:-compress_config.locality_saving[layer_idx],:],
                    compress_config.quantize_bit[layer_idx],
                    128,
                )
            )
    if compress_config.compress_method[layer_idx] == "uniformquantization":
        # print("begin uniquant",starting_idx,locality_idx,compress_config.quantize_bit[layer_idx])
        previous_key[:, :, starting_idx:-locality_idx, :] = fake_uniformquantization(
            previous_key[:, :, starting_idx:-locality_idx, :],
            compress_config.quantize_bit[layer_idx],
        )
        if previous_value is not None:
            previous_value[:, :, starting_idx:-locality_idx, :] = (
                fake_uniformquantization(
                    previous_value[:, :, starting_idx:-locality_idx, :],
                    compress_config.quantize_bit[layer_idx],
                )
            )
    if compress_config.compress_method[layer_idx] == "poweriteration":
        previous_key = fake_poweriteration(
            previous_key,
            compress_config.loop[layer_idx],
            compress_config.rank[layer_idx],
            compress_config.device_num[layer_idx],
            pbase1,
            qbase1,
        )
        if previous_value is not None:
            previous_value = fake_poweriteration(
                previous_value,
                compress_config.loop[layer_idx],
                compress_config.rank[layer_idx],
                compress_config.device_num[layer_idx],
                pbase2,
                qbase2,
            )
        pass
    if compress_config.compress_method[layer_idx] == "outlierquantization":
        # print("seqlen:",seq_len,"starting_idx:",starting_idx,"locality_idx:",locality_idx)
        previous_key[:, :, starting_idx:-locality_idx, :] = (
            fake_dense_sparse_uniformquantization(
                previous_key[:, :, starting_idx:-locality_idx, :],
                compress_config.quantize_bit[layer_idx],
                compress_config.left[layer_idx],
            )
        )
        if previous_value is not None:
            previous_value[:, :, starting_idx:-locality_idx, :] = (
                fake_dense_sparse_uniformquantization(
                    previous_value[:, :, starting_idx:-locality_idx, :],
                    compress_config.quantize_bit[layer_idx],
                    compress_config.left[layer_idx],
                )
            )
    if compress_config.compress_method[layer_idx] == "quantize_with_lrap":
        smaller_dim = seq_len if seq_len <= num_head * sep_dim else num_head * sep_dim
        smaller_dim = int(smaller_dim)
        rank_k = int(smaller_dim * compress_config.rank[layer_idx])
        rank_v = int(smaller_dim * compress_config.rankv[layer_idx])
        previous_key = fake_quant_with_lrap(
            previous_key,
            compress_config.quantize_bit[layer_idx],
            rank_k,
            compress_config.loop[layer_idx],
        )
        if previous_value is not None:
            previous_value = fake_quant_with_lrap(
                previous_value,
                compress_config.quantize_bit[layer_idx],
                rank_v,
                compress_config.loop[layer_idx],
            )
    if compress_config.compress_method[layer_idx] == "outliterquantize_with_lrap":
        smaller_dim = seq_len if seq_len <= num_head * sep_dim else num_head * sep_dim
        smaller_dim = int(smaller_dim)
        rank_k = int(smaller_dim * compress_config.rank[layer_idx])
        rank_v = int(smaller_dim * compress_config.rankv[layer_idx])
        previous_key = fake_outquant_with_lrap(
            previous_key,
            compress_config.quantize_bit[layer_idx],
            rank_k,
            compress_config.loop[layer_idx],
            compress_config.left[layer_idx],
        )
        if previous_value is not None:
            previous_value = fake_outquant_with_lrap(
                previous_value,
                compress_config.quantize_bit[layer_idx],
                rank_v,
                compress_config.loop[layer_idx],
                compress_config.left[layer_idx],
            )
    # TODO take a look at this
    if compress_config.compress_method[layer_idx] == "outquantize_with_lrap_iter":
        smaller_dim = seq_len if seq_len <= num_head * sep_dim else num_head * sep_dim
        smaller_dim = int(smaller_dim)
        rank_k = int(smaller_dim * compress_config.rank[layer_idx])
        rank_v = int(smaller_dim * compress_config.rankv[layer_idx])
        previous_key = fake_outquant_with_lrap_iter(
            previous_key,
            compress_config.quantize_bit[layer_idx],
            rank_k,
            compress_config.loop[layer_idx],
            compress_config.left[layer_idx],
            compress_config.iter[layer_idx],
        )
        if previous_value is not None:
            previous_value = fake_outquant_with_lrap_iter(
                previous_value,
                compress_config.quantize_bit[layer_idx],
                rank_v,
                compress_config.loop[layer_idx],
                compress_config.left[layer_idx],
                compress_config.iter[layer_idx],
            )
    if compress_config.compress_method[layer_idx] == "outliterquantize_with_lrap_preserving":
        smaller_dim = seq_len if seq_len <= num_head * sep_dim else num_head * sep_dim
        smaller_dim = int(smaller_dim)
        rank_k = int(smaller_dim * compress_config.rank[layer_idx])
        rank_v = int(smaller_dim * compress_config.rankv[layer_idx])
        # [bsz, num_head, seq_len, sep_dim]
        
        previous_key[:,:,compress_config.start_saving[layer_idx]:-compress_config.locality_saving[layer_idx],:] = fake_outquant_with_lrap(
            previous_key[:,:,compress_config.start_saving[layer_idx]:-compress_config.locality_saving[layer_idx],:],
            compress_config.quantize_bit[layer_idx],
            rank_k,
            compress_config.loop[layer_idx],
            compress_config.left[layer_idx],
        )
        if previous_value is not None:
            previous_value[:,:,compress_config.start_saving[layer_idx]:-compress_config.locality_saving[layer_idx],:] = fake_outquant_with_lrap(
                previous_value[:,:,compress_config.start_saving[layer_idx]:-compress_config.locality_saving[layer_idx],:],
                compress_config.quantize_bit[layer_idx],
                rank_v,
                compress_config.loop[layer_idx],
                compress_config.left[layer_idx],
            )

    # SK: TODO take a look at this: idea merged from KIVI
    if compress_config.compress_method[layer_idx] == "group_channel_with_lrap":
        smaller_dim = seq_len if seq_len <= num_head * sep_dim else num_head * sep_dim
        smaller_dim = int(smaller_dim)
        rank_k = int(smaller_dim * compress_config.rank[layer_idx])
        rank_v = int(smaller_dim * compress_config.rankv[layer_idx])
        previous_key = fake_group_channel_quant_with_lrap(
            previous_key,
            compress_config.quantize_bit[layer_idx],
            rank_k,
            compress_config.loop[layer_idx],
        )
        if previous_value is not None:
            previous_value = fake_group_channel_quant_with_lrap(
                previous_value,
                compress_config.quantize_bit[layer_idx],
                rank_v,
                compress_config.loop[layer_idx],
            )

    # SK: TODO take a look at this: idea merged from KIVI
    if compress_config.compress_method[layer_idx] == "group_token_with_lrap":
        smaller_dim = seq_len if seq_len <= num_head * sep_dim else num_head * sep_dim
        smaller_dim = int(smaller_dim)
        rank_k = int(smaller_dim * compress_config.rank[layer_idx])
        rank_v = int(smaller_dim * compress_config.rankv[layer_idx])
        previous_key = fake_group_token_quant_with_lrap(
            previous_key,
            compress_config.quantize_bit[layer_idx],
            rank_k,
            compress_config.loop[layer_idx],
        )
        if previous_value is not None:
            previous_value = fake_group_token_quant_with_lrap(
                previous_value,
                compress_config.quantize_bit[layer_idx],
                rank_v,
                compress_config.loop[layer_idx],
            )

    # SK: TODO take a look at this: idea merged from KIVI
    if compress_config.compress_method[layer_idx] == "group_kc_vt_with_lrap":
        smaller_dim = seq_len if seq_len <= num_head * sep_dim else num_head * sep_dim
        smaller_dim = int(smaller_dim)
        rank_k = int(smaller_dim * compress_config.rank[layer_idx])
        rank_v = int(smaller_dim * compress_config.rankv[layer_idx])
        previous_key = fake_group_channel_quant_with_lrap(
            previous_key,
            compress_config.quantize_bit[layer_idx],
            rank_k,
            compress_config.loop[layer_idx],
        )
        if previous_value is not None:
            previous_value = fake_group_token_quant_with_lrap(
                previous_value,
                compress_config.quantize_bit[layer_idx],
                rank_v,
                compress_config.loop[layer_idx],
            )
    return previous_key, previous_value


# iteratively compress the input tensor and simluate the error by low rank approximation
def fake_outquant_with_lrap_iter(tensor, quantize_bit, rank, loop, left, iter):
    lrap_error = tensor
    batch, num_head, seq_len, sep_dim = tensor.shape
    p_base = [torch.rand(sep_dim * num_head, rank).to(tensor.device)]
    q_base = [torch.rand(batch * seq_len, rank).to(tensor.device)]
    for i in range(iter):
        tensor_quantized = fake_dense_sparse_uniformquantization(
            lrap_error, quantize_bit, left
        )
        tensor_error = tensor - tensor_quantized
        tensor_error_lrap = fake_poweriteration(
            tensor_error, loop, rank, tensor_quantized.device, p_base, q_base
        )
        lrap_error = tensor - tensor_error_lrap
    tensor_return = tensor_quantized + tensor_error_lrap
    return tensor_return


def fake_outquant_with_lrap(tensor, quantize_bit, rank, loop, left):
    tensor_quantized = fake_dense_sparse_uniformquantization(tensor, quantize_bit, left)
    tensor_error = tensor - tensor_quantized
    tensor_error_lrap = fake_poweriteration(
        tensor_error, loop, rank, tensor_quantized.device, None, None
    )
    tensor_return = tensor_quantized + tensor_error_lrap
    return tensor_return


def fake_quant_with_lrap(tensor, quantize_bit, rank, loop):
    tensor_quantized = fake_uniformquantization(tensor, quantize_bit)
    tensor_error = tensor - tensor_quantized
    tensor_error_lrap = fake_poweriteration(
        tensor_error, loop, rank, tensor_quantized.device, None, None
    )
    tensor_return = tensor_quantized + tensor_error_lrap
    return tensor_return


def fake_group_channel_quant_with_lrap(tensor, quantize_bit, rank, loop):
    tensor_quantized = fake_groupwise_channel_asymmetric_quantization(
        tensor, quantize_bit, 128
    )
    tensor_error = tensor - tensor_quantized
    tensor_error_lrap = fake_poweriteration(
        tensor_error, loop, rank, tensor_quantized.device, None, None
    )
    tensor_return = tensor_quantized + tensor_error_lrap
    return tensor_return


def fake_group_token_quant_with_lrap(tensor, quantize_bit, rank, loop):
    tensor_quantized = fake_groupwise_token_asymmetric_quantization(
        tensor, quantize_bit, 128
    )
    tensor_error = tensor - tensor_quantized
    tensor_error_lrap = fake_poweriteration(
        tensor_error, loop, rank, tensor_quantized.device, None, None
    )
    tensor_return = tensor_quantized + tensor_error_lrap
    return tensor_return


# def fake_top_k_pruning(input_tensor, rate):
#     shape = input_tensor.shape
#     input_tensor = input_tensor.reshape(-1)
#     map_tensor = input_tensor.abs()
#     map_tensor = map_tensor.float()  # convert to 32bits to avoid max - min = inf
#     sortedtensor, indices = torch.sort(map_tensor)
#     k = int(len(sortedtensor) * rate)
#     threshold = sortedtensor[-k]
#     input_tensor[input_tensor < threshold] = 0
#     input_tensor = input_tensor.reshape(shape)
#     return input_tensor
